// cohort.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2023-12-15 11:19:09 tac>
//
// 年齢等の時期項目別コホート分析を行う。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W cohort cohort.C -I/usr/local/include
//               -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "cohort  Version 7.42         Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:cohort  [<オプション> …]  [<データファイル> …]",
  "機能:最初の項目を時期項目、以後を集計項目としたコホート分析をするんや。",
  "<オプション>",
  "  -dt=<時期間隔>:時間項目の時間間隔を指定するんや。(オプション省略時:-dt=1)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込むんや。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するんや。(-h:-h=man)",
  "  -ngp={msd|mse}[,…]:Ngraph用出力データを指定するんや。",
  "    msd:平均値±SD  mse:平均値±SE  オプション省略時:全て指定",
  "  -ngpname[=<出力ファイル名>]:Ngraph用データの出力ファイル名を指定するんや。",
  "    オプション省略時:Ngraph用データ出力せず  -ngpname:-ngpname=cohort_ngp.dat",
  "  -res=[-][{def|all|statc|statt}][,…]:解析結果の出力項目を指定するんや。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    statc:コホート別基礎統計量  statt:全体の基礎統計量",
  "  -t=<時間1>,<時間2>[,<時間3>,…]:項目の相対時間を指定するんや。",
  "    オプション省略時:-t=0,1,…,(項目数-1)"
};

const string Title = "=== コホート分析(cohort analysys) ==="; // 表題
const string NGPNAMEdef = "cohort_ngp.dat"; // 省略時Ngraphファイル名

using vecstatv = vector<statv<double>>; // 基礎統計量ベクトルの短名定義

enum ngptype                    // Ngraph用出力データの種類
{
  ngpnon,                       // 出力なし
  ngpmsd,                       // 平均値±SD
  ngpmse,                       // 平均値±SE
  ngpetc                        // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vdt;                   // 時期間隔
  string vngpname;              // Ngraphファイル基本名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  lstr tlst;                    // 項目相対時間リスト
  vecdbl vtvec;                 // 項目相対時間ベクトル
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double dt() const             // 時期間隔を返す
  { return vdt; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t(size_t idx) const    // 項目時間を返す
  { return (idx < vtvec.size()) ? vtvec[idx] : BLANKf; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vdt(1.0), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), tlst(), vtvec()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dt")           // 時期間隔の指定
  {
    vdt = strtodf(opt[1]);
    if (vdt <= 0.0 || vdt == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "t")       // 項目時間の指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni < 2) return disperr(vpname, "項目が2つ以上必要でっせ！");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  vtvec.resize(ni);
  for (size_t j = 0; j < ni; ++j) vtvec[j] = (double)j * vdt;
  if (settvec(vtvec, tlst, ni) < 0) return disperr(vpname, "項目相対時間が間違っとるんや！");
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["msd"] = vngpmap["mse"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "-ngpnameオプションの指定が必要！");
    vngpmap["msd"] = vngpmap["mse"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "-ngp=" + ngp + "オプション指定が不適切！"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["statc"] = vresmap["statt"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["statc"] = vresmap["statt"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "-res=" + res + "オプション指定が不適切！"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["statc"] = vresmap["statt"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "-res=" + res + "オプション指定が不適切！"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

using mapdv = map<double, vecstatv>; // コホート別統計量マップの別名定義
using mapdv_itr = mapdv::iterator; // 反復子の別名定義
using mapdv_citr = mapdv::const_iterator; // const反復子の別名定義

class cohortclass               // cohortメインクラス
{
private:
  string fname;                 // ファイル名
  void vecstatv_sigma(vecstatv& statvec, datac& cdata); // データを累積する
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_chlist(ostream& fout, option& opt, mapdv& vecstatvmap) const;
  // コホート表を出力する
  void sum_yt(const option& opt, double tmin, vecstatv& ytvec, mapdv& vecstatvmap) const;
  // コホートを加算する
  void plot_ngp(option& opt, idd& id, mapdv& vecstatvmap) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_msde(ofstream& fout, const option& opt, ngptype ngp, mapdv& vecstatvmap) const;
  // Ngraph用に平均値±SD/SEを出力する
public:
  cohortclass() : fname()       // コンストラクタ
  {}
  ~cohortclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み累積する
  void put(option& opt, idd& id, ldata& datalst);
  // 集計結果を出力する
};

void cohortclass::vecstatv_sigma(vecstatv& statvec, datac& cdata)
// データを累積する
{
  int idxend = min(statvec.size() + 1, cdata.n());
  for (int i = 1; i < idxend; ++i)
  {
    double fdata = cdata.fdata(i);
    if (fdata != BLANKf) statvec[i - 1].sigma(fdata);
  }
}

void cohortclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "コホート項目:" << id.item(0) << "\n";
  fout << "時期項目   1:" << id.item(1) << "\n";
  fout << "  :\n";
  fout << "時期項目" << setw(4) << id.ni() - 1;
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void cohortclass::put_chlist(ostream& fout, option& opt, mapdv& vecstatvmap) const
// コホート表を出力する
{
  constexpr int LENline = 74;   // 区切り線の長さ
  double tmin = 0.0, tmax = 0;
  int k = 0;
  for (mapdv_itr chitr = vecstatvmap.begin(); chitr != vecstatvmap.end(); ++chitr)
  {
    int nd = 0;
    for (size_t j = 0; j < chitr->second.size(); ++j)
    {
      if ((chitr->second[j]).n() > 0) ++nd;
    }
    if (nd > 0)
    {
      ++k;
      if (k == 1) tmin = chitr->first;
      if (opt.res("statc"))
      {
        fout << "\n・コホート" << k << ":開始時期t=" << chitr->first << "\n";
        putline(fout, LENline);
      }
      double t = 0.0;
      for (size_t j = 0; j < chitr->second.size(); ++j)
      {
        statv<double>* yj = &(chitr->second[j]);
        if (yj->n() > 0)
        {
          t = chitr->first + opt.t(j);
          if (opt.res("statc"))
          {
            fout << "t=" << setw(4) << t << ":";
            putmsd(fout, yj->n(), yj->mean(), yj->vr());
          }
        }
      }
      if (opt.res("statc")) putline(fout, LENline);
      if (t > tmax) tmax = t;
    }
  }
  if (opt.res("statt") == false) return;
  int p = int(tmax - tmin + 0.99) + 1;
  vecstatv ytvec(p);
  sum_yt(opt, tmin, ytvec, vecstatvmap);
  fout << "\n・全体:開始時期t=" << tmin << "\n";
  putline(fout, LENline);
  for (int i = 0; i < p; ++i)
  {
    if (ytvec[i].n() > 0)
    {
      fout << "t=" << setw(4) << tmin + (double)i * opt.dt() << ":";
      putmsd(fout, ytvec[i].n(), ytvec[i].mean(), ytvec[i].vr());
    }
  }
  putline(fout, LENline);
}

void cohortclass::sum_yt(const option& opt, double tmin, vecstatv& ytvec, mapdv& vecstatvmap) const
// コホートを加算する
{
  for (mapdv_itr chitr = vecstatvmap.begin(); chitr != vecstatvmap.end(); ++chitr)
  {
    for (size_t j = 0; j < chitr->second.size(); ++j)
    {
      if ((chitr->second[j]).n() > 0)
      {
        int idx = (int)(chitr->first + opt.t(j) - tmin);
        ytvec[idx] += chitr->second[j];
      }
    }
  }
}

void cohortclass::plot_ngp(option& opt, idd& id, mapdv& vecstatvmap) const
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("mse"))           // 平均値±SE
  {
    fout << "-s1 -dline -w80\n";
    plot_ngpid(fout, id);
    plot_msde(fout, opt, ngpmse, vecstatvmap);
    delml = true;
  }
  if (opt.ngp("msd"))           // 平均値±SD
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80\n";
    plot_ngpid(fout, id);
    plot_msde(fout, opt, ngpmsd, vecstatvmap);
  }
}

void cohortclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%SD・SEを描く時は-derrorbar_y -w40 -o500\n";
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%コホート項目:" << id.item(0) << "\n";
  fout << "%時期項目   1:" << id.item(1) << "\n";
  fout << "%  :\n";
  fout << "%時期項目" << setw(4) << id.ni() - 1;
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void cohortclass::plot_msde(ofstream& fout, const option& opt, ngptype ngp, mapdv& vecstatvmap) const
// Ngraph用平均値±SD/SE出力
{
  fout << "%平均値±";
  if (ngp == ngpmsd) fout << "SD\n";
  else fout << "SE\n";
  double tmin = 0.0;
  int k = 0;
  for (mapdv_itr chitr = vecstatvmap.begin(); chitr != vecstatvmap.end(); ++chitr)
  {
    if (k > 0) fout << "= =\n";
    fout << "%コホート" << k + 1 << "\n";
    int nd = 0;
    for (size_t j = 0; j < chitr->second.size(); ++j)
    {
      if ((chitr->second[j]).n() > 0) ++nd;
    }
    if (nd > 0)
    {
      ++k;
      if (k == 1) tmin = chitr->first;
      double t = 0.0;
      for (size_t j = 0; j < chitr->second.size(); ++j)
      {
        statv<double>* yj = &(chitr->second[j]);
        if (yj->n() > 0)
        {
          t = chitr->first + opt.t(j);
          double ebar = (ngp == ngpmsd) ? yj->sd() : yj->se();
          fout << t << " " << yj->mean() << " " << -ebar << " " << ebar << endl;
        }
      }
    }
  }
}

int cohortclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー！");
  return nc;
}

void cohortclass::put(option& opt, idd& id, ldata& datalst)
// 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      cout << "\n" << endl;
    }
    return;
  }
  mapdv vecstatvmap;            // コホート別統計量ベクトルマップ
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double t = ditr->fdata(0);
    if (t != BLANKf)
    {
      mapdv_itr chitr = vecstatvmap.find(t);
      if (chitr == vecstatvmap.end()) // 新規コホート
      {
        vecstatv statvec(ditr->n() - 1);
        for (size_t i = 1; i < ditr->n(); ++i)
        {
          double fdata = ditr->fdata(i);
          if (fdata != BLANKf) statvec[i - 1].sigma(fdata);
        }
        vecstatvmap.insert(map<double, vecstatv>::value_type(t, statvec));
      }
      else                      // 既存のコホート
      { vecstatv_sigma(chitr->second, *ditr); }
    }
  }
  if (opt.res("res"))           // 計算結果出力
  {
    put_resid(cout, id);
    put_chlist(cout, opt, vecstatvmap);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(opt, id, vecstatvmap);
}

int main(int argc, char* argv[]) // メインルーチン
{
  cohortclass cohort;           // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) cohort.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { cohort.get(argv[i], opt, id, datalst);  fok = false; }
    }
    cohort.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
