// fft.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2023-12-15 11:22:02 tac>
//
// 高速フーリエ変換による周期解析を行う。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W fft fft.C -I/usr/local/include
//               -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H,mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "fft  Version 7.42            Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:fft  [<オプション> …]  [<データファイル> …]",
  "機能:高速フーリエ変換による周期解析をするでござる。",
  "<オプション>",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込むでござる。",
  "  -h[=<コマンド名>]:オンラインマニュアルの表示でござる。(-h:-h=man)",
  "  -k=<番号>[,<番号>,…]:検定に組み込む周期成分番号を指定するでござる。",
  "    オプション省略時:-k=1,2",
  "  -ngp={lsp|rsp}[,…]:Ngraph用出力データを指定するでござる。",
  "    lsp:線スペクトル  rsp:寄与率スペクトル  オプション省略時:全て指定",
  "  -ngpname[=<出力ファイル名>]:Ngraph用データの出力ファイル名を指定する。",
  "    オプション省略時:Ngraph用データ出力せず  -ngpname:-ngpname=fft_ngp.dat",
  "  -pm={0|2}:周期解析の手法を指定するでござる。",
  "    0:平均値だけを用いる単純法  2:二元配置法(オプション省略時)",
  "  -res=[-][{def|all|cc[<信頼係数>]|pctab[a]|stat}][,…]",
  "      :解析結果の出力項目を指定するでござる。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    cc[<信頼係数>]:信頼区間(0≦信頼係数＜1)",
  "    pctab[a]:周期成分表(a:全ての周期成分を出力)  stat:基礎統計量",
  "  -T=<基本周期>:基本周期(1周期の時間)を指定するでござる。",
  "    オプション省略時:-T=24"
};

const string Title = "=== フーリエ解析(Fourier analysis) ==="; // 表題
const string NGPNAMEdef = "fft_ngp.dat"; // 省略時Ngraphファイル名
constexpr double T0def = 24.0;    // 省略時基本周期

enum pctabtype                  // 周期成分表出力の種類
{
  pctabnon,                     // 出力せず
  pctabauto,                    // 選択した周期成分だけ出力
  pctaball,                     // 全て出力
  pctabetc                      // その他
};

enum pmtype                     // 周期解析の種類
{
  pmnon,                        // なし
  pm0,                          // 平均値のみ
  pm1,                          // 一元配置型(未使用)
  pm2,                          // 二元配置型
  pmetc                         // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  lstr klst;                    // 周期成分番号リスト
  vecint vpcvec;                // 周期成分番号配列
  string vngpname;              // Ngraphファイル基本名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  pctabtype vpctab;             // 周期成分表の出力フラッグ
  pmtype vpm;                   // 周期解析の種類
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  double vt0;                   // 基本周期
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  size_t m() const              // 周期成分数を返す
  { return vpcvec.size(); }
  int pcn(size_t k) const       // 周期成分番号を返す
  { return (k < vpcvec.size()) ? vpcvec[k] : 0; }
  pctabtype pctab() const       // 周期成分表出力フラッグを返す
  { return vpctab; }
  pmtype pm() const             // 周期解析の種類を返す
  { return vpm; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t0() const             // 基本周期を返す
  { return vt0; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), klst(), vpcvec(), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vpctab(pctabauto), vpm(pm2), vt0(T0def)
// コンストラクタ
{
  if (c_pname == nullptr) return;     // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "k")       // 周期成分番号の指定
  {
    if (strsplit(klst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pm")      // 周期解析手法指定
  {
    if (opt[1].empty() || opt[1] == "2") vpm = pm2;
    else if (opt[1] == "0") vpm = pm0;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "T")       // 基本周期の指定
  {
    vt0 = strtodf(opt[1]);
    if (vt0 == BLANKf || vt0 <= 0.0) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni < 3) return disperr(vpname, "時期項目が3つ以上必要でござる！");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  int q = ni / 2;               // 周期成分最大数
  if (klst.empty())             // 周期成分指定がない時
  {
    vpcvec.resize(2);
    vpcvec[0] = 1;  vpcvec[1] = 2;
  }
  else if (setpcvec(vpcvec, klst, q) < 0)
  { return disperr(vpname, "周期成分指定が間違ってござる！"); }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["lsp"] = vngpmap["rsp"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "-ngpnameオプションの指定が必要！");
    vngpmap["lsp"] = vngpmap["rsp"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "-ngp=" + ngp + "オプション指定が不適切！"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "-res=" + res + "オプション指定が不適切！"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = true;
        vpctab = pctaball;      // 全ての周期成分を出力する
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        vresmap["cc"] = true;
      }
      else if (res == "pctaba")
      { vresmap["pctab"] = true;  vpctab = pctaball; }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "-res=" + res + "オプション指定が不適切！"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  int np;                        // 時期数
  vector<statv<double>> stabvec; // 時期項目の基本統計量ベクトル
  double vsb;                    // 観察個体平方和
  double vsa;                    // 時期平方和
  int vdfa;                      // 時期自由度
  double vssp;                   // 周期平方和
  double vst;                    // 総平方和
  pfun<double> vpf0;             // 全体周期関数オブジェクト
  pfun<double> vpf;              // 選択周期関数オブジェクト
  int sigma_fdata(const option& opt, datac& cdata);
  // 時期項目のデータを累積する
  int cal_fft( vecdbl& abvec, const option& opt);
  // 基礎統計の計算とFFT
public:
  varclass(size_t ni) : np(ni), stabvec(ni + 1), vsb(0.0), vsa(0.0), vdfa(0), vssp(0.0), vst(0.0), vpf0(), vpf() // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int n(size_t idx) const       // 時期項目の例数を返す
  { return (idx < stabvec.size()) ? stabvec[idx].n() : 0; }
  double mean(size_t idx) const // 時期項目の平均値を返す
  { return (idx < stabvec.size()) ? stabvec[idx].mean() : 0; }
  double vr(size_t idx) const   // 時期項目の不偏分散を返す
  { return (idx < stabvec.size()) ? stabvec[idx].vr() : 0; }
  double sb() const             // 観察個体平方和を返す
  { return vsb; }
  double sa() const             // 時期平方和を返す
  { return vsa; }
  int dfa() const               // 時期自由度を返す
  { return vdfa; }
  double ssp() const            // 周期平方和を返す
  { return vssp; }
  double st() const             // 総平方和を返す
  { return vst; }
  const pfun<double> pf0() const // 全体周期関数オブジェクトを返す
  { return vpf0; }
  const pfun<double> pf() const // 選択周期関数オブジェクトを返す
  { return vpf; }
  int cal(const option& opt, ldata& datalst); // データを集計してFFT計算をする
};

int varclass::sigma_fdata(const option& opt, datac& cdata)
// 時期項目のデータを累積する
{
  vecdbl fdvec(np);
  bool blank = false;
  for (int j = 0; j < np; ++j)
  {
    if ((fdvec[j] = cdata.fdata(j)) == BLANKf) blank = true;
  }
  if (blank && opt.pm() == pm2) return 0; // 二元配置法で欠損のある時は集計せず
  double sum = 0.0;
  for (int j = 0; j < np; ++j)
  {
    if (fdvec[j] != BLANKf) { stabvec[j].sigma(fdvec[j]);  sum += fdvec[j]; }
  }
  vsb += sum * sum;
  return 1;                     // 累積した
}

int varclass::cal_fft(vecdbl& abvec, const option& opt)
// 基礎統計の計算とFFT
{
  int p = np;
  int q = p / 2;
  double w = 2.0 * M_PI / (double)p;
  vsa = 0.0;
  for (int j = 0; j < p + 2; ++j) abvec[j] = 0.0;
  bool blank = false;
  for (int j = 0; blank == false && j < p; ++j)
  {
    int n = stabvec[j].n();
    if (n < 1)
    { blank = true; }
    else
    {
      stabvec[p] += stabvec[j];
      double m = stabvec[j].mean();
      for (int k = 1; k <= q; ++k)
      {
        int i = 2 * k;
        double tk = w * (double)(k * j);
        abvec[i - 1] += m * cos(tk);
        if (i < p) abvec[i] += m * sin(tk);
      }
      abvec[0] += m;  abvec[p + 1] += m * m;  vsa += n * m * m;
    }
  }
  if (blank) return -1;         // 欠損あり！
  for (int k = 1; k <= q; ++k)
  {
    int i = 2 * k;
    if (i == p) abvec[i - 1] /= (double)(p);
    else { abvec[i - 1] /= (double)p * 0.5;  abvec[i] /= (double)p * 0.5; }
  }
  abvec[p + 1] -= abvec[0] * abvec[0] / (double)p;
  abvec[0] /= (double)p;
  vst = (opt.pm() == pm2) ? stabvec[p].ss() : abvec[p + 1];
  double cf = stabvec[p].n() * stabvec[p].mean() * stabvec[p].mean();
  const double Seps = DBL_EPSILON * fmax(5.0, vst);
  vsa -= cf;  vsb = vsb / (double)p - cf;
  if (vsa < Seps) vsa = 0.0;
  if (vsb < Seps) vsb = 0.0;
  return q;                     // 周期成分数を返す
}

int varclass::cal(const option& opt, ldata& datalst) // データを集計してFFT計算をする
{
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr) sigma_fdata(opt, *ditr);
  int p = np;
  vecdbl abvec0(p + 2);
  if (cal_fft(abvec0, opt) < 0) return -1; // 欠損あり！
  vssp = abvec0[p + 1];
  int q = p / 2;
  vecint pcvec0(q);
  for (int k = 0; k < q; ++k) pcvec0[k] = k + 1;
  pfun<double> pf0(opt.t0(), pcvec0, abvec0);
  pcvec0.resize(opt.m());
  vecdbl abvec(opt.m() * 2 + 2);
  abvec[0] = abvec0[0];
  for (size_t k = 0; k < opt.m(); ++k)
  {
    pcvec0[k] = opt.pcn(k);
    int i = (k + 1) * 2, j = opt.pcn(k) * 2;
    abvec[i - 1] = abvec0[j - 1];
    abvec[i] = abvec0[j];
  }
  abvec[opt.m() * 2 + 1] = abvec0[p + 1];
  pfun<double> pf(opt.t0(), pcvec0, abvec);
  vdfa = p - 1;  vpf0 = pf0;  vpf = pf;
  return q;                     // 周期成分数を返す
}

class fftclass                  // fftメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 時期別基礎統計量を出力する
  void put_fftfun(ostream& fout, const option& opt, const idd& id, const pfun<double>& pf, double ssp) const;
  // FFT用周期関数パラメターを出力する
  void put_ffttab(ostream& fout, option& opt, const idd& id, const varclass& var) const;
  // フーリエ解析用分散分析表を出力する
  void put_ffttab0(ostream& fout, const option& opt, const idd& id, const pfun<double>& pf, double sp, int dfp, double sr, int dfr, double st, int dft) const;
  // FFT用周期回帰分散分析表を出力する
  void put_ffttab2(ostream& fout, const option& opt, const idd& id, const pfun<double>& pf, double sp, int dfp, double sb, double sl, int dfl, double sa, int dfa, int dfb, double sr, int dfr, double st, int dft) const;
  // 二元配置型FFT用周期回帰分散分析表を出力する
  void plot_ngp(option& opt, idd& id, const pfun<double>& pf, double ssp) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_linespect(ofstream& fout, const idd& id, const pfun<double>& pf) const;
  // Ngraph用に線スペクトルデータを出力する
  void plot_r2spect(ofstream& fout, const idd& id, const pfun<double>& pf, double ssp) const;
  // Ngraph用に寄与率スペクトルデータを出力する
public:
  fftclass() : fname()          // コンストラクタ
  {}
  ~fftclass()                   // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void fftclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "時期項目  1:" << id.item(0) << "\n";
  fout << "  :\n";
  fout << "時期項目" << setw(3) << id.ni();
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void fftclass::put_stat(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 時期別基礎統計量を出力する
{
  constexpr int LENline = 77;   // 区切り線の長さ
  fout << "\n時間 t   :基礎統計量\n";
  putline(fout, LENline);
  int p = id.ni();
  double dt = opt.t0() / (double)p;
  for (int j = 0; j < p; ++j)
  {
    int prec0 = fout.precision(2); // 小数点以下の精度を指定する
    fout.setf(ios::fixed);
    fout << setw(9) << dt * (double)j << setprecision(prec0); // 精度を元に戻す
    fout.unsetf(ios::fixed);
    fout << ":";  putmsd(fout, var.n(j), var.mean(j), var.vr(j));
  }
  putline(fout, LENline);
  fout << "全体     :";  putmsd(fout, var.n(p), var.mean(p), var.vr(p));
}

void fftclass::put_fftfun(ostream& fout, const option& opt, const idd& id, const pfun<double>& pf, double ssp) const
// FFT用周期関数パラメターを出力する
{
  fout << "\n・周期回帰式(periodic regression equation):y = A0 + ΣAk・cos(ωk・t - θk)\n";
  fout << "    k=1,…,p:周期成分  A0:メサー(定数)  Ak:振幅  ωk=360°・k/T:角周波数\n";
  fout << "    θk:位相  T=" << pf.t(0) << ":基本周期  T/k:周期  f=1/T:基本周波数  fk=k/T:周波数\n";
  if (opt.pctab() == pctabnon) return;
  fout << "\n周期成分      振幅  位相(度)  位相時間            回帰係数      寄与率    周期\n";
  fout << "   k            Ak      θk°  θk/ωk          ak          bk      r2     T/k\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  fout << "   0" << setw(14) << pf.amp(0) << "(メサー)\n";
  int p = id.ni();
  double deg = 180.0 / M_PI;
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    int pcn = pf.pcn(k);
    double dfk = (2 * pcn < p) ? 2.0 : 1.0;
    double r2 = pf.amp(k) * pf.amp(k) * (double)p / dfk / ssp;
    fout << setw(4) << pcn << setw(14) << pf.amp(k);
    fout << setw(10) << pf.acr(k) * deg;
    fout << setw(10) << pf.acr(k) / pf.w(k);
    fout << setw(12) << pf.a(k) << setw(12) << pf.b(k);
    int prec0 = fout.precision(5);
    fout.setf(ios::fixed);
    fout << setw(8) << r2;
    fout.unsetf(ios::fixed);
    fout << setw(8) << pf.t(k) << setprecision(prec0) << "\n";
  }
  putline(fout, LENline);
}

void fftclass::put_ffttab(ostream& fout, option& opt, const idd& id, const varclass& var) const
// フーリエ解析用分散分析表を出力する
{
  if (var.st() <= 0.0) return;
  int p = (int)(id.ni());
  int dft = (opt.pm() == pm0) ? var.dfa() : var.n(p) - 1;
  double sp = 0.0;
  int dfp = 0;
  for (size_t k = 1; k <= opt.m(); ++k)
  {
    int pcn = (var.pf()).pcn(k);
    int dfk = (2 * pcn < p) ? 2 : 1;
    dfp += dfk;
    sp += (var.pf()).amp(k) * (var.pf()).amp(k) * (double)p / (double)dfk;
  }
  int nc = var.n(0);
  if (opt.pm() == pm2) sp *= (double)nc;
  double sl = var.sa() - sp;
  const double Seps = DBL_EPSILON * fmax(5.0, var.st());
  if (sl < Seps) sl = 0.0;
  double sr = var.st() - sp;
  int dfl = var.dfa() - dfp, dfb = nc - 1, dfr = dft - dfp;
  if (opt.pm() == pm2)
  { sr = var.st() - var.sa() - var.sb();  dfr = dft - var.dfa() - dfb; }
  if (sr < Seps) sr = 0.0;
  double v = (dfr > 0) ? sr / (double)dfr / (double)(dft + 1) : 0.0;
  double ci = calmcci(mcfis, opt.cc(), 1, dfp, dfr, 1, v);
  if (opt.res("cc"))
  {
    fout << "・メサーの" << opt.cc() * 100.0 << "%信頼区間=";
    fout << (var.pf()).amp(0) << "±" << ci << "(" << (var.pf()).amp(0) - ci << "-" << (var.pf()).amp(0) + ci << ")\n";
  }
  if (opt.pm() == pm2) put_ffttab2(fout, opt, id, var.pf(), sp, dfp, var.sb(), sl, dfl, var.sa(), var.dfa(), dfb, sr, dfr, var.st(), dft);
  else put_ffttab0(fout, opt, id, var.pf(), sp, dfp, sr, dfr, var.st(), dft);
}

void fftclass::put_ffttab0(ostream& fout, const option& opt, const idd& id, const pfun<double>& pf, double sp, int dfp, double sr, int dfr, double st, int dft) const
// FFT用周期回帰分散分析表を出力する
{
  string tstr("周期分散分析表(periodic ANOVA table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << "\n" << setw(58 + npad) << tstr;
  fout << "\n";
  fout << "要因              平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  int np = id.ni();
  double v = 0.0, f = 0.0, p = 0.0;
  for (size_t k = 1; k <= opt.m();  ++k)
  {
    int pcn = pf.pcn(k);
    int dfk = (2 * pcn < np) ? 2 : 1;
    double spk = pf.amp(k) * pf.amp(k) * (double)(dft + 1) / (double)dfk;
    calvfp(v, f, p, spk, dfk, sr, dfr);
    fout << "周期成分" << setw(3) << pcn << setw(13) << spk << setw(10) << dfk;
    fout << setw(14) << v << setw(12) << f << setw(14) << p << pind(p) << "\n";
  }
  putline(fout, LENline);
  double vr = calvfp(v, f, p, sp, dfp, sr, dfr);
  tstr = "全周期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
  fout << setw(10) << dfp << setw(14) << v << setw(12) << f;
  fout << setw(14) << p << pind(p) << "\n";
  tstr = "残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr << "\n";
  putline(fout, LENline);
  tstr = "全体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft << endl;
}

void fftclass::put_ffttab2(ostream& fout, const option& opt, const idd& id, const pfun<double>& pf, double sp, int dfp, double sb, double sl, int dfl, double sa, int dfa, int dfb, double sr, int dfr, double st, int dft) const
// 二元配置型FFT用周期回帰分散分析表を出力する
{
  string tstr("二元配置型周期分散分析表(periodic ANOVA table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << "\n" << setw(62 + npad) << tstr;
  fout << "\n";
  fout << "要因              平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  int np = id.ni();
  double v = 0.0, f = 0.0, p = 0.0;
  for (size_t k = 1; k <= opt.m();  ++k)
  {
    int pcn = pf.pcn(k);
    int dfk = (2 * pcn < np) ? 2 : 1;
    double spk = pf.amp(k) * pf.amp(k) * (double)(dft + 1) / (double)dfk;
    calvfp(v, f, p, spk, dfk, sr, dfr);
    fout << "周期成分" << setw(3) << pcn << setw(13) << spk << setw(10) << dfk;
    fout << setw(14) << v << setw(12) << f << setw(14) << p << pind(p) << "\n";
  }
  putline(fout, LENline);
  double vr = calvfp(v, f, p, sp, dfp, sr, dfr);
  tstr = "全周期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
  fout << setw(10) << dfp << setw(14) << v << setw(12) << f;
  fout << setw(14) << p << pind(p) << "\n";
  calvfp(v, f, p, sl, dfl, sr, dfr);
  tstr = "ズレ";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
  fout << setw(10) << dfl << setw(14) << v << setw(12) << f;
  fout << setw(14) << p << pind(p) << "\n";
  putline(fout, LENline);
  calvfp(v, f, p, sa, dfa, sr, dfr);
  tstr = "時期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f;
  fout << setw(14) << p << pind(p) << "\n";
  calvfp(v, f, p, sb, dfb, sr, dfr);
  tstr = "個体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sb;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f;
  fout << setw(14) << p << pind(p) << "\n";
  tstr = "残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr << "\n";
  putline(fout, LENline);
  tstr = "全体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft << endl;
}

void fftclass::plot_ngp(option& opt, idd& id, const pfun<double>& pf, double ssp) const
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("lsp"))           // 線スペクトル
  {
    fout << "-s1 -drectangle_solid_fill -x1 -y3 -c255,0,0 -zx0,";
    fout << id.ni() / 2 + 1 << ",1\n";
    plot_ngpid(fout, id);
    plot_linespect(fout, id, pf);
    delml = true;
  }
  if (opt.ngp("rsp"))           // 寄与率スペクトル
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -drectangle_solid_fill -x1 -y3 -c255,0,0 -zx0,";
      fout << id.ni() / 2 + 1 << ",1\n";
    }
    plot_ngpid(fout, id);
    plot_r2spect(fout, id, pf, ssp);
  }
}

void fftclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title); fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%時期項目  1:" << id.item(0) << "\n";
  fout << "%  :\n";
  fout << "%時期項目" << setw(3) << id.ni();
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void fftclass::plot_linespect(ofstream& fout, const idd& id, const pfun<double>& pf) const
// Ngraph用に線スペクトルデータを出力する
{
  fout << "%線スペクトル\n";
  int p = id.ni();
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    double dfk = (2 * pf.pcn(k) < p) ? 2.0 : 1.0;
    double apow = pf.amp(k) * pf.amp(k) / dfk;
    fout << (double)k - 0.05 << " 0 " << (double)k + 0.05 << " " << apow << endl;
  }
}

void fftclass::plot_r2spect(ofstream& fout, const idd& id, const pfun<double>& pf, double ssp) const
// Ngraph用に寄与率スペクトルデータを出力する
{
  fout << "%寄与率スペクトル\n";
  int p = id.ni(), m = pf.m();
  for (int k = 1; k <= m; ++k)
  {
    double dfk = (2 * pf.pcn(k) < p) ? 2.0 : 1.0;
    double r2 = pf.amp(k) * pf.amp(k) * (double)p / dfk / ssp;
    fout << (double)k - 0.05 << " 0 " << (double)k + 0.05 << " " << r2 << endl;
  }
}

int fftclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー！");
  return nc;
}

void fftclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (id.nc() < 1)              // データ無し
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var(id.ni());        // 変数オブジェクト
  if (var.cal(opt, datalst) < 0)
  {
    disperr(opt.pname(), "欠損あり！");
    return;
  }
  if (opt.res("res"))
  {
    put_resid(cout, id);
    if (opt.res("stat")) put_stat(cout, opt, id, var);
    if (opt.res("pctab"))
    {
      if (opt.pctab() == pctaball) put_fftfun(cout, opt, id, var.pf0(), var.ssp());
      else put_fftfun(cout, opt, id, var.pf(), var.ssp());
      put_ffttab(cout, opt, id, var);
    }
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(opt, id, var.pf0(), var.ssp());
}

int main(int argc, char* argv[]) // メインルーチン
{
  fftclass fft;                 // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理モードフラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) fft.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { fft.get(argv[i], opt, id, datalst);  fok = false; }
    }
    fft.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
