// fftc.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2024-01-14 15:27:19 tac>
//
// 高速フーリエ変換によるケース方向の周期解析を行う。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W fftc.C -I/usr/local/include -L/usr/local/lib
//               -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H,mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "fftc  Version 7.42           Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:fftc  [<オプション> …]  [<データファイル> …]",
  "機能:高速フーリエ変換によるケース方向の周期解析をするでござる。",
  "<オプション>",
  "  -dt[=<データの時間間隔>]:データの時間間隔を指定するでござる。",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込むでござる。",
  "  -h[=<コマンド名>]:オンラインマニュアルの表示でござる。(-h:-h=man)",
  "  -k=<番号>[,<番号>,…]:検定に組み込む周期成分番号を指定するでござる。",
  "    オプション省略時:-k=1,2",
  "  -ngp={data|prc|psd|rsd}[,…]:Ngraph用出力データを指定するでござる。",
  "    data:生データ  prc:周期回帰曲線  psd:パワースペクトル",
  "    rsd:寄与率スペクトル",
  "  -ngpname[=<出力ファイル名>]:Ngraph用データの出力ファイル名を指定する。",
  "    オプション省略時:Ngraph用データ出力せず  -ngpname:-ngpname=fftc_ngp.dat",
  "  -res=[-][{def|all|cc[<信頼係数>]|pctab[a]|stat}][,…]",
  "      :解析結果の出力項目を指定するでござる。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    cc[<信頼係数>]:信頼区間(0≦信頼係数＜1)",
  "    pctab[a]:周期成分表(a:全ての周期成分を出力)  stat:基礎統計量",
  "  -T=<基本周期>:基本周期(1周期の時間)を指定するでござる。",
  "    オプション省略時:-T=24"
};

const string Title = "=== フーリエ解析(Fourier analysis) ==="; // 表題
const string NGPNAMEdef = "fftc_ngp.dat"; // 省略時Ngraphファイル名
constexpr double T0def = 24.0;    // 省略時基本周期

using mapsb = map<string, bool>; // boolマップの別名定義
using mapsb_itr = mapsb::iterator;        // 反復子の別名定義
using mapsb_citr = mapsb::const_iterator; // const反復子の別名定義

using lstdbl = list<double>;    // 倍精度実数リストの別名定義


enum pctabtype                  // 周期成分表出力の種類
{
  pctabnon,                     // 出力せず
  pctabauto,                    // 選択した周期成分だけ出力
  pctaball,                     // 全て出力
  pctabetc                      // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  double vdt;                   // データの時間間隔
  lstr klst;                    // 周期成分番号リスト
  vecint vpcvec;                // 周期成分番号配列
  string vngpname;              // Ngraphファイル基本名
  lstr vngplst;                 // Ngraph出力リスト
  mapsb vngpmap;                // Ngraph出力の有無マップ
  pctabtype vpctab;             // 周期成分表の出力フラッグ
  lstr vreslst;                 // 分析結果出力リスト
  mapsb vresmap;                // 分析結果出力の有無マップ
  double vt0;                   // 基本周期
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double dt() const             // データの時間間隔
  { return vdt; }
  int m() const                 // 周期成分数を返す
  { return vpcvec.size(); }
  int pcn(size_t k) const       // 周期成分番号を返す
  { return (k < vpcvec.size()) ? vpcvec[k] : 0; }
  pctabtype pctab() const       // 周期成分表出力フラッグを返す
  { return vpctab; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t0() const             // 基本周期を返す
  { return vt0; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int p);             // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdt(0.0), klst(), vpcvec(), vngpname(), vngplst(), vngpmap(), vpctab(pctabauto), vreslst(), vresmap(), vt0(T0def)
// コンストラクタ
{
  if (c_pname == nullptr) return;     // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  mapsb_itr ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  mapsb_itr resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dt")           // データの時間間隔の指定
  {
    vdt = strtodf(opt[1]);
    if (vdt == BLANKf || vdt <= 0.0) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "k")       // 周期成分番号の指定
  {
    if (strsplit(klst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "T")       // 基本周期の指定
  {
    vt0 = strtodf(opt[1]);
    if (vt0 == BLANKf || vt0 <= 0.0) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(int p)        // オプションのチェックとセット
{
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  int q = p / 2;                // 周期成分最大数
  if (vdt <= 0.0) vdt = vt0 / (double)p;
  if (klst.empty())             // 周期成分指定がない時
  {
    vpcvec.resize(2);
    vpcvec[0] = 1;  vpcvec[1] = 2;
  }
  else if (setpcvec(vpcvec, klst, q) < 0)
  { return disperr(vpname, "周期成分指定が間違ってござる！"); }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["data"] = vngpmap["prc"] = vngpmap["psd"] = vngpmap["rsd"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "-ngpnameオプションの指定が必要！");
    vngpmap["data"] = vngpmap["prc"] = vngpmap["psd"] = vngpmap["rsd"] = false;
    for (string ngp : vngplst)
    {
      mapsb_itr ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "-ngp=" + ngp + "オプション指定が不適切！"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      mapsb_itr resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "-res=" + res + "オプション指定が不適切！"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = true;
      }
      else if (res == "all")   // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["pctab"] = vresmap["stat"] = true;
        vpctab = pctaball;      // 全ての周期成分を出力する
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        vresmap["cc"] = true;
      }
      else if (res == "pctaba")
      { vresmap["pctab"] = true;  vpctab = pctaball; }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "-res=" + res + "オプション指定が不適切！"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  lstdbl vylst;                 // 測定データのリスト
  int vnp;                      // データ数
  double vmean;                 // 平均値
  double vst;                   // 平方和
  pfun<double> vpf0;            // 全体周期関数オブジェクト
  pfun<double> vpf;             // 選択周期関数オブジェクト
  void cal_fftc(vecdbl& abvec, const option& opt); // 基礎統計の計算とFFT
public:
  varclass() : vylst(), vnp(0), vmean(0.0), vst(0.0), vpf0(), vpf() // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  const lstdbl& ylst() const    // 測定データのリストを返す
  { return vylst; }
  int np() const                // データ数を返す
  { return vnp; }
  double mean() const           // 平均値を返す
  { return vmean; }
  double st() const             // 平方和を返す
  { return vst; }
  const pfun<double> pf0() const // 全体周期関数オブジェクトを返す
  { return vpf0; }
  const pfun<double> pf() const // 選択周期関数オブジェクトを返す
  { return vpf; }
  int cal(option& opt, ldata& datalst); // データを集計してFFT計算をする
};

void varclass::cal_fftc(vecdbl& abvec, const option& opt)
// 基礎統計の計算とFFT
{
  int p = vnp;
  int q = p / 2;
  // const double wt = opt.dt() * 2.0 * M_PI / (double)p;
  const double w = 2.0 * M_PI / (double)p;
  abvec.assign(p + 2, 0.0);
  int j = 0;
  for (double y : vylst)
  {
    for (int k = 1; k <= q; ++k)
    {
      int i = 2 * k;
      double tk = w * (double)(k * j);
      abvec[i - 1] += y * cos(tk);
      if (i < p) abvec[i] += y * sin(tk);
    }
    abvec[0] += y;  abvec[p + 1] += y * y;
    ++j;
  }
  for (int k = 1; k <= q; ++k)
  {
    int i = 2 * k;
    if (i == p) abvec[i - 1] /= (double)p;
    else { abvec[i - 1] /= (double)p * 0.5;  abvec[i] /= (double)p * 0.5; }
  }
  abvec[p + 1] -= abvec[0] * abvec[0] / (double)p; // データの平方和(St)
  const double ABeps = DBL_EPSILON * fmax(5.0, w);
  if (abvec[p + 1] < ABeps) abvec[p + 1] = 0.0;
  abvec[0] /= (double)p;        // データの平均(My)
}

int varclass::cal(option& opt, ldata& datalst) // データを集計してFFT計算をする
{
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double y = ditr->fdata(0);
    if (y != BLANKf) vylst.push_back(y);
  }
  if (opt.check(vylst.size()) != 0) return disperr(opt.pname(), "実行時オプションエラー！");
  vnp = vylst.size();
  int p = vnp;
  vecdbl abvec0(p + 2);
  cal_fftc(abvec0, opt);
  vmean = abvec0[0];  vst = abvec0[p + 1];
  int q = p / 2;
  vecint pcvec0(q, 0);
  for (int k = 0; k < q; ++k) pcvec0[k] = k + 1;
  pfun pf0(opt.t0(), pcvec0, abvec0);
  pcvec0.resize(opt.m());
  vecdbl abvec(opt.m() * 2 + 2);
  abvec[0] = abvec0[0];
  for (int k = 0; k < opt.m(); ++k)
  {
    pcvec0[k] = opt.pcn(k);
    int i = (k + 1) * 2, j = opt.pcn(k) * 2;
    abvec[i - 1] = abvec0[j - 1];
    abvec[i] = abvec0[j];
  }
  abvec[opt.m() * 2 + 1] = abvec0[p + 1];
  pfun pf(opt.t0(), pcvec0, abvec);
  vpf0 = pf0;  vpf = pf;
  return q;                     // 周期成分数を返す
}

class fftcclass                 // fftcメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, const option& opt, const varclass& var) const;
  // 時期別基礎統計量を出力する
  void put_fftcfun(ostream& fout, const option& opt, int p, const pfun<double>& pf, double st) const;
  // FFT用周期関数パラメターを出力する
  void put_fftctab(ostream& fout, option& opt, int p, const pfun<double>& pf, double st) const;
  // フーリエ解析用分散分析表を出力する
  void put_fftctab0(ostream& fout, int p, const pfun<double>& pf, double sp, int dfp, double sr, int dfr, double st, int dft) const;
  // FFT用周期回帰分散分析表を出力する
  void plot_ngp(option& opt, idd& id, const varclass& var) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpdata(ofstream& fout, const option& opt, const lstdbl& ylst) const;
  // Ngraph用に生データを出力する
  void plot_ngpprc(ofstream& fout, const pfun<double>& pf) const;
  // Ngraph用に周期回帰曲線データを出力する
  void plot_ngppsd(ofstream& fout, const option& opt, int p, const pfun<double>& pf) const;
  // Ngraph用にパワースペクトルデータを出力する
  void plot_ngprsd(ofstream& fout, const option& opt, int p, const pfun<double>& pf, double st) const;
  // Ngraph用に寄与率スペクトルデータを出力する
public:
  fftcclass() : fname()         // コンストラクタ
  {}
  ~fftcclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void fftcclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "データ項目:" << id.item(0) << "\n";
}

void fftcclass::put_stat(ostream& fout, const option& opt, const varclass& var) const
// データの基礎統計量を出力する
{
  int p = var.np();
  double vr = (p > 1) ? var.st() / (p - 1) : 0.0;
  fout << "基礎統計量:";  putmsd(fout, p, var.mean(), vr);
}

void fftcclass::put_fftcfun(ostream& fout, const option& opt, int p, const pfun<double>& pf, double st) const
// FFT用周期関数パラメターを出力する
{
  fout << "\n・周期回帰式(periodic regression equation):y = A0 + ΣAk・cos(ωk・t - θk)\n";
  fout << "    k=1,…,p:周期成分  A0:メサー(定数)  Ak:振幅  ωk=360°・k/T:角周波数\n";
  fout << "    θk:位相  T=" << pf.t(0) << ":基本周期  T/k:周期  f=1/T:基本周波数  fk=k/T:周波数\n";
  fout << "    T/p=" << opt.dt() << ":時間間隔\n";
  if (opt.pctab() == pctabnon) return;
  fout << "\n周期成分      振幅  位相(度)  位相時間            回帰係数      寄与率    周期\n";
  fout << "   k            Ak      θk°  θk/ωk          ak          bk      r2     T/k\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  fout << "   0" << setw(14) << pf.amp(0) << "(メサー)\n";
  double deg = 180.0 / M_PI;
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    double dfk = (2 * pf.pcn(k) < p) ? 2.0 : 1.0;
    double r2 = pf.amp(k) * pf.amp(k) * (double)p / dfk / st;
    fout << setw(4) << pf.pcn(k) << setw(14) << pf.amp(k);
    fout << setw(10) << pf.acr(k) * deg;
    fout << setw(10) << pf.acr(k) / pf.w(k);
    fout << setw(12) << pf.a(k) << setw(12) << pf.b(k);
    int prec0 = fout.precision(5); // 小数点以下の精度を変更する
    fout.setf(ios::fixed);
    fout << setw(8) << r2;
    fout.unsetf(ios::fixed);
    fout << setw(8) << pf.t(k) << setprecision(prec0) << "\n";
  }
  putline(fout, LENline);
}

void fftcclass::put_fftctab(ostream& fout, option& opt, int p, const pfun<double>& pf, double st) const
// フーリエ解析用分散分析表を出力する
{
  if (st <= 0.0) return;
  int dft = p - 1;
  double sp = 0.0;
  int dfp = 0;
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    int dfk = (2 * pf.pcn(k) < p) ? 2 : 1;
    dfp += dfk;
    sp += pf.amp(k) * pf.amp(k) * (double)p / (double)dfk;
  }
  const double Seps = DBL_EPSILON * fmax(5.0, st);
  if (sp < Seps) sp = 0.0;
  double sr = st - sp;
  if (sr < Seps) sr = 0.0;
  int dfr = dft - dfp;
  double v = (dfr > 0) ? sr / (double)dfr / (double)(dft + 1) : 0.0;
  double ci = calmcci(mcfis, opt.cc(), 1, dfp, dfr, 1, v);
  if (opt.res("cc"))
  {
    fout << "・メサーの" << opt.cc() * 100.0 << "%信頼区間=";
    fout << pf.amp(0) << "±" << ci << "(" << pf.amp(0) - ci << "-" << pf.amp(0) + ci << ")\n";
  }
  put_fftctab0(fout, p, pf, sp, dfp, sr, dfr, st, dft);
}

void fftcclass::put_fftctab0(ostream& fout, int p, const pfun<double>& pf, double sp, int dfp, double sr, int dfr, double st, int dft) const
// FFT用周期回帰分散分析表を出力する
{
  string tstr("周期分散分析表(periodic ANOVA table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << "\n" << setw(58 + npad) << tstr;
  fout << "\n";
  fout << "要因              平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    int dfk = (2 * pf.pcn(k) < p) ? 2 : 1;
    double spk = pf.amp(k) * pf.amp(k) * (double)(dft + 1) / (double)dfk;
    calvfp(v, f, pr, spk, dfk, sr, dfr);
    fout << "周期成分" << setw(3) << pf.pcn(k) << setw(13) << spk << setw(10) << dfk;
    fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  }
  putline(fout, LENline);
  double vr = calvfp(v, f, pr, sp, dfp, sr, dfr);
  tstr = "全周期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
  fout << setw(10) << dfp << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  tstr = "残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr << "\n";
  putline(fout, LENline);
  tstr = "全体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft << endl;
}

void fftcclass::plot_ngp(option& opt, idd& id, const varclass& var) const
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("data"))          // 生データ
  {
    fout << "-s1 -dmark,00 -o300\n";
    plot_ngpid(fout, id);
    plot_ngpdata(fout, opt, var.ylst());
    delml = true;
  }
  if (opt.ngp("prc"))           // 周期回帰曲線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -c255,0,0\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_ngpprc(fout, var.pf());
  }
  if (opt.ngp("psd"))           // パワースペクトル
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -c255,0,0\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_ngppsd(fout, opt, var.np(), var.pf0());
  }
  if (opt.ngp("rsd"))           // 寄与率スペクトル
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -c255,0,0\n";
    plot_ngpid(fout, id);
    plot_ngprsd(fout, opt, var.np(), var.pf0(), var.st());
  }
}

void fftcclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%データ項目:" << id.item(0) << "\n";
}

void fftcclass::plot_ngpdata(ofstream& fout, const option& opt, const lstdbl& ylst) const
// Ngraph用に個々のデータを出力する
{
  fout << "%生データ\n";
  double t = 0.0;
  for (double y : ylst)
  {
    fout << t << " " << y << endl;
    t += opt.dt();
  }
}

void fftcclass::plot_ngpprc(ofstream& fout, const pfun<double>& pf) const
// Ngraph用周期回帰曲線を出力する
{
  fout << "%周期回帰曲線\n";
  constexpr uint RDdef = 60;    // 曲線の分解数
  double dt = pf.t(0) / (double)RDdef;
  double t = 0.0;
  for (uint i = 0; i <= RDdef; ++i, t += dt)
  { fout << t << " " << pf.fun(t) << endl; }
}

void fftcclass::plot_ngppsd(ofstream& fout, const option& opt, int p, const pfun<double>& pf) const
// Ngraph用にパワースペクトルデータを出力する
{
  fout << "%パワースペクトル\n";
  double ff = 1.0 / opt.t0();   // 周期成分番号->周波数変換ファクター
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    double dfk = (2 * pf.pcn(k) < p) ? 2.0 : 1.0;
    double apow = pf.amp(k) * pf.amp(k) / dfk;
    fout << (double)k * ff << " " << apow << endl;
  }
}

void fftcclass::plot_ngprsd(ofstream& fout, const option& opt, int p, const pfun<double>& pf, double st) const
// Ngraph用に寄与率スペクトルデータを出力する
{
  fout << "%寄与率スペクトル\n";
  double ff = 1.0 / opt.t0();   // 周期成分番号->周波数変換ファクター
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    double dfk = (2 * pf.pcn(k) < p) ? 2.0 : 1.0;
    double r2 = pf.amp(k) * pf.amp(k) * (double)p / dfk / st;
    fout << (double)k * ff << " " << r2 << endl;
  }
}

int fftcclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (id.ni() > 1) throw runtime_error("項目は測定データだけでござる！");
  return nc;
}

void fftcclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (id.nc() < 1)              // データ無し
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var;                 // 変数オブジェクト
  if (var.cal(opt, datalst) < 0) return;
  if (opt.res("res"))
  {
    put_resid(cout, id);
    if (opt.res("stat")) put_stat(cout, opt, var);
    if (opt.res("pctab"))
    {
      if (opt.pctab() == pctaball) put_fftcfun(cout, opt, var.np(), var.pf0(), var.st());
      else put_fftcfun(cout, opt, var.np(), var.pf(), var.st());
      put_fftctab(cout, opt, var.np(), var.pf(), var.st());
    }
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(opt, id, var);
}

int main(int argc, char* argv[]) // メインルーチン
{
  fftcclass fftc;               // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理モードフラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) fftc.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { fftc.get(argv[i], opt, id, datalst);  fok = false; }
    }
    fftc.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
