// pcova.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2024-03-12 09:28:03 tac>
//
// 最初の項目を群項目、以後の項目を時期項目とした周期共分散分析を行う。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W pcova pcova.C -I/usr/local/include
//               -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H,mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "pcova  Version 7.42          Copyleft(Vo\\oV) 2024 杉本典夫",
  "書式:pcova [<オプション> …]  [<データファイル> …]",
  "機能:最初の項目を群項目、以後の項目を時期項目として周期共分散分析を行う。",
  "<オプション>",
  "  -dt=<時期間隔>:時間項目の時間間隔を指定する。(オプション省略時:-dt=T/項目数)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込む。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示する。(-h:-h=man)",
  "  -k=<番号>[,<番号>,…]:周期回帰式に組み込む周期成分番号を指定する。",
  "  -pm={0|1|2}:周期共分散分析の手法を指定する。",
  "    0:平均値を用いる単純法  1:一元配置法  2:二元配置法(オプション省略時)",
  "  -res=[-][{def|all|atab|cc[<信頼係数>]|m[{d|m}][{b|f|s|t}]|r|statg|statt}][,…]",
  "      :解析結果の出力項目を指定するでござる。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    atab:分散分析表  cc[<信頼係数>]:信頼区間(0≦信頼係数＜1  cc:cc0.95)",
  "    m[{d|m}][{b|f|s|t}]:多重比較(md:非平行性  mm:メサー)",
  "      b:Bonferroni型  f:Fisher型  s:Scheffe型  t:Tukey型",
  "    r:相関係数と一致係数  statg:群別基礎統計量  statt:全体の基礎統計量",
  "  -T=<基本周期>:基本周期(1周期の時間)を指定する。",
  "  -t=<時間1>,<時間2>[,<時間3>,…]:時期項目の時間を指定する。"
};

const string Title = "=== 周期共分散分析(periodic analysis of covariance) ==="; // 表題
constexpr double T0def = 24.0;  // 省略時基本周期

using vecstatv = vector<statv<double>>; // 基礎統計量ベクトルの別名定義

enum pmtype                     // 周期解析の種類
{
  pmnon,                        // なし
  pm0,                          // 平均値を用いる単純法
  pm1,                          // 一元配置法
  pm2,                          // 二元配置法
  pm3,                          // 群にも対応のある二元配置法
  pmetc                         // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  double vdt;                   // 時期間隔
  lstr klst;                    // 周期成分番号リスト
  vecint vpcvec;                // 周期成分番号配列
  int vpcmax;                   // 周期成分番号最大値
  mctype vmd;                   // 非平行性の多重比較法
  mctype vmm;                   // メサーの多重比較法
  pmtype vpm;                   // 周期解析の種類
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  double vt0;                   // 基本周期
  lstr tlst;                    // 項目時間リスト
  vecdbl vtvec;                 // 項目時間ベクトル
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double dt() const             // 時期間隔を返す
  { return vdt; }
  int m() const                 // 周期成分数を返す
  { return vpcvec.size(); }
  mctype md() const             // 非平行性の多重比較法を返す
  { return vmd; }
  mctype mm() const             // メサーの多重比較法を返す
  { return vmm; }
  const vecint& pcvec() const   // 周期成分番号ベクトルを返す
  { return vpcvec; }
  int pcn(size_t k) const       // 周期成分番号を返す
  { return (k < vpcvec.size()) ? vpcvec[k] : 0; }
  int pcmax() const             // 周期成分番号最大値を返す
  { return vpcmax; }
  pmtype pm() const             // 周期解析の種類を返す
  { return vpm; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t0() const             // 基本周期を返す
  { return vt0; }
  double t(size_t idx) const    // 項目時間を返す
  { return (idx < vtvec.size()) ? vtvec[idx] : vt0; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int ni);            // オプションをチェックする
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdt(0.0), klst(), vpcvec(), vpcmax(-1), vmd(mcbon), vmm(mctuk), vpm(pm2), vreslst(), vresmap(), vt0(T0def), tlst(), vtvec()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dt")           // 時期間隔の指定
  {
    vdt = strtodf(opt[1]);
    if (vdt <= 0.0 || vdt == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "k")       // 周期成分番号の指定
  {
    if (strsplit(klst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pm")      // 周期解析の手法指定
  {
    if (opt[1].empty() || opt[1] == "2") vpm = pm2;
    else if (opt[1] == "0") vpm = pm0;
    else if (opt[1] == "1") vpm = pm1;
    else if (opt[1] == "3") vpm = pm3;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "T")       // 基本周期の指定
  {
    vt0 = strtodf(opt[1]);
    if (vt0 == BLANKf || vt0 <= 0.0) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "t")       // 項目時間の指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(int ni)       // オプションをチェックする
{
  if (ni < 3) return disperr(vpname, "項目が3つ以上必要！");
  int p = ni - 1, q = (ni - 1) / 2; // 時期数と周期成分最大数
  vtvec.resize(p);
  double dt = (vdt > 0.0) ? vdt : vt0 / (double)p;
  for (int j = 0; j < p; ++j) vtvec[j] = (double)j * dt;
  if (settvec(vtvec, tlst, p) < 0) return disperr(vpname, "項目時間が間違っている！");
  if (klst.empty())             // 周期成分指定がない時
  {
    vpcvec.resize(2);
    vpcvec[0] = 1;  vpcvec[1] = 2;  vpcmax = 2;
  }
  else if ((vpcmax = setpcvec(vpcvec, klst, q)) < 0)
  { return disperr(vpname, "周期成分指定が間違っている！"); }
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["statg"] = vresmap["statt"] = true;
  vresmap["md"] = vresmap["mm"] = vresmap["r"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["statg"] = vresmap["statt"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "-res=" + res + "オプション指定が不適切！"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["statg"] = vresmap["statt"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["md"] = vresmap["mm"] = vresmap["r"] = vresmap["statg"] = vresmap["statt"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        vresmap["cc"] = true;
      }
      else if (res == "m")      // -res=mと指定された時
      { vresmap["md"] = vresmap["mm"] = true; }
      else if (res[0] == 'm' && res[1] == 'd' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmd = mcbon;
          break;
        case 'f' : vmd = mcfis;
          break;
        case 's' : vmd = mcsch;
          break;
        default : return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        }
        vresmap["md"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'm' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmm = mcbon;
          break;
        case 'f' : vmm = mcfis;
          break;
        case 's' : vmm = mcsch;
          break;
        case 't' : vmm = mctuk;
          break;
        default : return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        }
        vresmap["mm"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "-res=" + res + "オプション指定が不適切！"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class gstat                     // 群別統計量クラス
{
private:
  vecstatv vystatvec;           // データの基礎統計量ベクトル
  int vpp;                      // データが存在する時期数
  int vnp;                      // 回帰用全例数
  mat<double> smat;             // 周期回帰計算用積和行列
  mat<double> pmat;             // 周期回帰計算後の残差行列
  double vsb;                   // 周期回帰平方和
  double vsp;                   // 時期変動平方和(回帰用Syy)
  pfun<double> vpf;             // x周期関数
public:
  gstat() : vystatvec(), vpp(-1), vnp(0), smat(), pmat(), vsp(0.0), vsb(0.0), vpf()
  // デフォールトコンストラクタ
  {}
  explicit gstat(int p0) : vystatvec(p0 + 1), vpp(-1), vnp(0), smat(), pmat(), vsp(0.0), vsb(0.0), vpf()
  // コンストラクタ
  {}
  gstat(const gstat& gstat0) : vystatvec(gstat0.vystatvec), vpp(gstat0.vpp), vnp(gstat0.vnp), smat(gstat0.smat), pmat(gstat0.pmat), vsp(gstat0.vsp), vsb(gstat0.vsb), vpf(gstat0.vpf)
  // コピーコンストラクタ
  {}
  virtual ~gstat()              // デストラクタ
  {}
  const gstat& operator=(const gstat& gstat0); // コピー代入演算子
  const gstat& operator+=(const gstat& gstat0); // 加算代入演算子
  double sigma(const vecdbl& fdatavec);         // データを累積する
  int pp();                     // データのある時期数を返す
  int np() const                // 回帰用全例数を返す
  { return vnp; }
  void setsmat(const option& opt); // 周期回帰計算用行列をセットする
  void calsmat(bool calm);         // 周期回帰用平均値を計算する
  void calpfun(const option& opt, bool calpf); // 周期関数を計算する
  statv<double>& ystat(int idx) // データの基礎統計量を返す
  { return vystatvec[idx]; }
  double ma(size_t idx) const   // cos項の平均を返す
  { return (0 < idx && idx <= vpf.m()) ? smat(0, idx * 2 - 1) : 0.0; }
  double mb(size_t idx) const   // sin項の平均を返す
  { return (idx <= vpf.m()) ? smat(0, idx * 2) : 0.0; }
  double mp() const             // 周期回帰用データの平均を返す
  { return smat(0, smat.c() - 1); }
  double sb() const             // 回帰平方和を返す
  { return vsb; }
  double sp() const             // 時期変動平方和(回帰用Syy)を返す
  { return vsp; }
  int m() const                 // 周期成分数を返す
  { return vpf.m(); }
  int pcn(int k) const          // 周期成分番号を返す
  { return vpf.pcn(k); }
  double a(int k) const         // cos項の回帰係数aを返す
  { return vpf.a(k); }
  double b(int k) const         // sin項の回帰係数bを返す
  { return vpf.b(k); }
  double amp(int k) const       // 振幅を返す
  { return vpf.amp(k); }
  double acr(int k) const       // 位相を返す
  { return vpf.acr(k); }
  double w(int k) const         // 角周波数を返す
  { return vpf.w(k); }
  double t(int k) const         // 周期を返す
  { return vpf.t(k); }
  double r2() const             // 周期回帰式の重寄与率を返す
  { return (vsp > 0.0) ? vsb / vsp : 0.0; }
  double r() const              // 周期回帰式の重相関係数を返す
  { return (vsp > 0.0) ? sqrt(vsb / vsp) : 0.0; }
  double vxr(const vec<double>& dxv) const; // 多重比較用dV^(-1)dを返す
  friend double pfinp(const gstat& gstata, const gstat& gstatb);
  // 周期関数の内積を返す
  friend double pfdinp(const gstat& gstata, const gstat& gstatb);
  // 周期関数の差の内積を返す
};

const gstat& gstat::operator=(const gstat& gstat0) // コピー代入演算子
{
  if (&gstat0 == this) return *this; // 自分自身への代入を防ぐ
  vystatvec = gstat0.vystatvec;  vpp = gstat0.vpp;  vnp = gstat0.vnp;
  smat = gstat0.smat;  pmat = gstat0.pmat;  vsp = gstat0.vsp;
  vsb = gstat0.vsb;  vpf = gstat0.vpf;
  return *this;                 // 自分自身を返す
}

const gstat& gstat::operator+=(const gstat& gstat0) // 加算代入演算子
{
  for (size_t j = 0; j < vystatvec.size(); ++j) vystatvec[j] += gstat0.vystatvec[j];
  if (smat.r() > 0)
  { smat += gstat0.smat;  pmat += gstat0.pmat; }
  else
  {
    smat.resize(gstat0.smat.r(), gstat0.smat.c());
    smat = gstat0.smat;
    pmat.resize(gstat0.pmat.r(), gstat0.pmat.c());
    pmat = gstat0.pmat;
  }
  return *this;                 // 自分自身を返す
}

double gstat::sigma(const vecdbl& fdatavec) // データを累積する
{
  int n = 0;
  double sum = 0.0;
  int p = vystatvec.size() - 1;
  for (int j = 0; j < p; ++j)
  {
    if (fdatavec[j] != BLANKf)
    { vystatvec[j].sigma(fdatavec[j]);  ++n;  sum += fdatavec[j]; }
  }
  return (n > 0) ? sum * sum / n : 0.0; // 個体平方和用の値を返す
}

int gstat::pp()                 // データのある時期数を返す
{
  if (vpp < 0)                  // vpp未集計
  {
    vpp = 0;
    int p = vystatvec.size() - 1;
    for (int j = 0; j < p; ++j)
    {
      if (vystatvec[j].n() > 0) vpp++;
    }
  }
  return vpp;
}

void gstat::setsmat(const option& opt) // 周期回帰式用行列をセットする
{
  int p = vystatvec.size() - 1, d = 2 * opt.m() + 1;
  smat.resize(d + 1, d + 1);
  pmat.resize(d + 1, d + 1);
  smat = 0.0;
  const double W = 2.0 * M_PI / opt.t0();
  int r = (opt.pm() == pm0) ? 1 : vystatvec[0].n();
  vecdbl xyvec(d + 1);
  xyvec[0] = 1.0;
  for (int j = 0; j < p; ++j)
  {
    if (vystatvec[j].n() > 0)
    {
      vystatvec[p] += vystatvec[j];  xyvec[d] = vystatvec[j].mean();
      for (int k = 0; k < opt.m(); ++k)
      {
        int i = (k + 1) * 2;
        double tk = W * opt.pcn(k) * opt.t(j);
        xyvec[i - 1] = cos(tk);
        xyvec[i] = (2 * opt.pcn(k) < p) ? sin(tk) : 0.0;
      }
      if (opt.pm() == pm1) r = vystatvec[j].n();
      for (int i = 0; i <= d; ++i)
      {
        double x = r * xyvec[i];
        for (int k = i; k <= d; ++k) smat.add(i, k, x * xyvec[k]);
      }
    }
  }
  pmat = smat;
}

void gstat::calsmat(bool calm)  // 周期回帰用平均値を計算する
{
  for (size_t i = 0; i < smat.r(); ++i)
  {
    for (size_t j = i + 1; j < smat.c(); ++j)
    { smat.assign(j, i, smat(i, j));  pmat.assign(j, i, pmat(i, j)); }
  }
  if (!calm) return;            // smatを計算しない(共通周期回帰用)
  vnp = (int)(smat(0, 0));
  smat.sweep(0);
  pmat = smat;
  vsp = smat(smat.r() - 1, smat.c() - 1);
  const double Veps = DBL_EPSILON * fmax(5.0, smat.maxabse());
  if (vsp < Veps) vsp = 0.0;
}

void gstat::calpfun(const option& opt, bool calpf) // 周期関数を計算する
{
  int d = pmat.r() - 1;
  vsb = pmat(d, d);
  vecdbl abvec(d);
  for (int i = 1; i < d; ++i) pmat.sweep(i);
  for (int i = 0; i < d; ++i) abvec[i] = pmat(i, d);
  vsb -= pmat(d, d);
  const double Veps = DBL_EPSILON * fmax(5.0, pmat.maxabse());
  if (vsb < Veps) vsb = 0.0;
  if (!calpf) return;           // pfunを計算しない(共通周期回帰用)
  pfun tpf(opt.t0(), opt.pcvec(), abvec);
  vpf = tpf;
}

double gstat::vxr(const vec<double>& dxv) const // 多重比較用dV^(-1)dを返す
{
  int d = dxv.dim();
  vec<double> tv = pmat.sub(1, 1, d, d) * dxv;
  return inp(tv, dxv);
}

double pfinp(const gstat& gstata, const gstat& gstatb)
// 周期関数の内積を返す
{
  int d = gstata.smat.r() - 1;
  vec<double> bvb(gstatb.pmat.sub(1, d, d - 1, d));
  mat<double> vmat(gstata.smat.sub(1, 1, d - 1, d - 1) + gstatb.smat.sub(1, 1, d - 1, d - 1));
  vmat *= 1.0 / (double)(gstata.np() + gstatb.np()); // 両者のVxxを合成する
  vec<double> tv(bvb.dim());
  tv = vmat * bvb;
  vec<double> tsubv(gstata.pmat.sub(1, d, d - 1, d));
  return inp(tsubv, tv);
}

double pfdinp(const gstat& gstata, const gstat& gstatb)
// 周期関数の差の内積を返す
{
  int d = gstata.smat.r() - 1;
  vec<double> dv = gstatb.pmat.sub(1, d, d - 1, d) - gstata.pmat.sub(1, d, d - 1, d);
  mat<double> vmat(gstata.smat.sub(1, 1, d - 1, d - 1) + gstatb.smat.sub(1, 1, d - 1, d - 1));
  vmat *= 1.0 / double(gstata.np() + gstatb.np()); // 両者のVxxを合成する
  vec<double> tv = vmat * dv;
  return inp(dv, tv);
}

using mapsg = map<string, gstat>;         // 群別統計量マップの別名定義
using mapsg_itr = mapsg::iterator;        // 反復子の別名定義
using mapsg_citr = mapsg::const_iterator; // const反復子の別名定義

class varclass                  // 変数クラス
{
private:
  int vnp;                      // 時期項目数
  mapsg vgsmap;                 // 群統計量マップ
  double vss;                   // 個体間平方和
  int vdfs;                     // 個体間自由度
  double vssr;                  // 個体残差平方和
  int vdfsr;                    // 個体残差自由度
  double vsa;                   // 群差平方和
  double vsaa;                  // 修正群差平方和
  int vdfa;                     // 群差自由度
  double vsbc;                  // 共通回帰平方和
  double vsb0;                  // 全体回帰平方和
  int vdfb;                     // 回帰自由度
  double vsd;                   // 非平行性平方和
  int vdfd;                     // 非平行性自由度
  double vsp;                   // 時期平方和
  int vdfp;                     // 時期自由度
  double vsl;                   // ズレ平方和
  int vdfl;                     // ズレ自由度
  double vsr;                   // 残差平方和
  int vdfr;                     // 残差自由度
  double vst;                   // 全体平方和
  int vdft;                     // 全体自由度
  gstat vgstatall;              // 全体の統計量オブジェクト
  gstat vgstatcom;              // 共通の統計量オブジェクト
  void cal_pcova(double& sa, int& dfa, double& ssr, int& dfsr, double& ss, int& dfs, double& sbc, double& sb0, int& dfb, double& saa, double& sd, int& dfd, double& sp, int& dfp, double& sl, int& dfl, double& sr, int& dfr, double& st, int& dft, const option& opt);
  // 周期共分散分析用各種パラメターを計算する
public:
  varclass(size_t ni) : vnp(ni - 1), vgsmap(), vss(0.0), vdfs(0), vssr(0.0), vdfsr(0), vsa(0.0), vsaa(0.0), vdfa(0), vsbc(0.0), vsb0(0.0), vdfb(0), vsd(0.0), vdfd(0), vsl(0.0), vdfl(0), vsr(0.0), vdfr(0), vst(0.0), vdft(0), vgstatall(ni - 1), vgstatcom(ni - 1) // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int np() const                // 時期項目数を返す
  { return vnp; }
  mapsg& gsmap()                // 群統計量マップを返す
  { return vgsmap; }
  int ng() const                // 群数を返す
  { return (int)(vgsmap.size()); }
  gstat& gstatall()             // 全体の統計量オブジェクトを返す
  { return vgstatall; }
  gstat& gstatcom()             // 共通の統計量オブジェクトを返す
  { return vgstatcom; }
  double ss() const             // 個体間平方和を返す
  { return vss; }
  int dfs() const               // 個体間自由度を返す
  { return vdfs; }
  double ssr() const            // 個体残差平方和を返す
  { return vssr; }
  int dfsr() const              // 個体残差自由度を返す
  { return vdfsr; }
  double sa() const             // 群差平方和を返す
  { return vsa; }
  double saa() const            // 修正群差平方和を返す
  { return vsaa; }
  int dfa() const               // 群差自由度を返す
  { return vdfa; }
  double sbc() const            // 共通回帰平方和を返す
  { return vsbc; }
  double sb0() const            // 全体回帰平方和を返す
  { return vsb0; }
  int dfb() const               // 回帰自由度を返す
  { return vdfb; }
  double sd() const             // 非平行性平方和を返す
  { return vsd; }
  int dfd() const               // 非平行性自由度を返す
  { return vdfd; }
  double sp() const             // 時期平方和を返す
  { return vsp; }
  int dfp() const               // 時期自由度を返す
  { return vdfp; }
  double sl() const             // ズレ平方和を返す
  { return vsl; }
  int dfl() const               // ズレ自由度を返す
  { return vdfl; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  double st() const             // 全体平方和を返す
  { return vst; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
  int cal(const option& opt, ldata& datalst); // データを集計して統計量を求める
  void pm3();                   // pm3用に個体残差を残差にする
};

void varclass::cal_pcova(double& sa, int& dfa, double& ssr, int& dfsr, double& ss, int& dfs, double& sbc, double& sb0, int& dfb, double& saa, double& sd, int& dfd, double& sp, int& dfp, double& sl, int& dfl, double& sr, int& dfr, double& st, int& dft, const option& opt)
// 周期共分散分析用各種パラメターを計算する
{
  int p = vnp, g = vgsmap.size();
  double ssp = 0.0, ssb = 0.0;
  dfp = 0;  sa = sl = 0.0;
  for (mapsg_itr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr)
  {
    dfp += gitr->second.pp();
    sa += gitr->second.np() * gitr->second.mp() * gitr->second.mp();
    ssp += gitr->second.sp();  ssb += gitr->second.sb();
    sl += gitr->second.sp() - gitr->second.sb();
  }
  const double Seps = DBL_EPSILON * fmax(5.0, ssp);
  if (sl < Seps) sl = 0.0;
  --dfp;  sb0 = vgstatall.sb();  sbc = vgstatcom.sb();
  dfb = 2 * opt.m();
  if (2 * opt.pcmax() >= p) --dfb;
  double cf = vgstatall.np() * vgstatall.mp() * vgstatall.mp();
  sa -= cf;  dfa = g - 1;  ss -= cf;  dfs = vgstatall.ystat(0).n() - 1;
  if (sa < Seps) sa = 0.0;
  if (ss < Seps) ss = 0.0;
  sp = sa + ssp;  saa = sa + sbc - sb0;  sd = ssb - sbc;
  if (sd < Seps) sd = 0.0;
  dfd = dfb * dfa;  dfl = g * ((p - 1) - dfb);
  if (dfl < 1) sl = 0.0;
  switch (opt.pm())
  {
  case pm0 : st = vgstatall.sp();  dft = vgstatall.np() - 1;
    ssr = sr = st - sa - ssb;  dfsr = dfr = dft - dfa - g * dfb;
    break;
  case pm1 : st = vgstatall.ystat(p).ss();  dft = vgstatall.ystat(p).n() - 1;
    ssr = sr = st - sp;  dfsr = dfr = dft - dfp;
    break;
  default : st = vgstatall.ystat(p).ss();  dft = vgstatall.ystat(p).n() - 1;
    ssr = ss - sa;  dfsr = dfs - dfa;
    sr = st - ss - ssp;  dfr = dft - dfs - g * (p - 1);
  }
  if (ssr < Seps) ssr = 0.0;
  if (sr < Seps) sr = 0.0;
}

int varclass::cal(const option& opt, ldata& datalst)
// データを集計して統計量を求める
{
  vecdbl fdatavec(vnp);
  vss = 0.0;
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    bool blank = false;
    for (int j = 0; j < vnp; ++j)
    {
      fdatavec[j] = ditr->fdata(j + 1);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (ditr->sdata(0) != BLANKs && (blank == false || opt.pm() == pm0 || opt.pm() == pm1))
    {
      mapsg_itr gitr = vgsmap.find(ditr->sdata(0));
      if (gitr == vgsmap.end())  // 新規群コード
      {
        gstat tgstat(fdatavec.size());
        vss += tgstat.sigma(fdatavec);
        vgsmap.insert(map<string, gstat>::value_type(ditr->sdata(0), tgstat));
      }
      else                      // すでにある群
      { vss += gitr->second.sigma(fdatavec); }
    }
  }
  if (vgsmap.empty()) return 0; // データなし
  int dfb = 2 * opt.m();
  if (2 * opt.pcmax() >= vnp) --dfb;
  for (mapsg_itr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr)
  {
      gitr->second.setsmat(opt);
      vgstatall += gitr->second;
      gitr->second.calsmat(true);
      vgstatcom += gitr->second;
      gitr->second.calpfun(opt, true);
  }
  vgstatall.calsmat(true);  vgstatcom.calsmat(false);
  vgstatall.calpfun(opt, true);  vgstatcom.calpfun(opt, true);
  cal_pcova(vsa, vdfa, vssr, vdfsr, vss, vdfs, vsbc, vsb0, vdfb, vsaa, vsd, vdfd, vsp, vdfp, vsl, vdfl, vsr, vdfr, vst, vdft, opt);
  return (int)(vgsmap.size());  // 群数を返す
}

void varclass::pm3()            // pm3用に個体残差を残差にする
{
  vssr = vsr;  vdfsr = vdfr;
}

class pcovaclass                // pcovaメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_mreg(ostream& fout, option& opt, const idd& id, varclass& var) const;
  // 平均・標準偏差・周期回帰式等を出力する
  void put_pfun(ostream& fout, const option& opt, const idd& id, gstat& gstati, gstat& gstatcom) const;
  // 周期関数パラメターを出力する
  void put_pcor(ostream& fout, varclass& var) const;
  // 周期関数の位相と振幅の一致係数を出力する
  void put_pcovatab0(ostream& fout, const varclass& var) const;
  // 周期共分散分析表を出力する
  void put_pcovatab1(ostream& fout, const varclass& var) const;
  // 一元配置型共分散分析表を出力する
  void put_pcovatab1rn(ostream& fout, double sa, int dfa, double sbc, double sb0, int dfb, double saa, double sd, int dfd, double sl, int dfl, double sp, int dfp, double sr, int dfr, double st, int dft) const;
  // 繰り返しの有る／無い一元配置型周期共分散分析表を出力する
  void put_pcovatab2(ostream& fout, const option& opt, const varclass& var) const;
  // 二元配置型周期共分散分析表を出力する
  void put_mmcmp(ostream& fout, const option& opt, varclass& var) const;
  // メサー(=修正群差)の多重比較を出力する
  void put_mmci(ostream& fout, const option& opt, varclass& var) const;
  // メサー(=修正群差)の信頼区間出力を出力する
  void put_mdcmp(ostream& fout, const option& opt, varclass& var) const;
  // 非平行性の多重比較を出力する
public:
  pcovaclass() : fname()        // コンストラクタ
  {}
  ~pcovaclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void pcovaclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "群項目     :" << id.item(0) << "\n";
  fout << "時期項目  1:" << id.item(1) << "\n";
  fout << "  :\n";
  fout << "時期項目" << setw(3) << id.ni() - 1;
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void pcovaclass::put_mreg(ostream& fout, option& opt, const idd& id, varclass& var) const
// 平均・標準偏差・周期回帰式等を出力する
{
  fout << "\n○周期回帰式(periodic regression equation)\n";
  fout << "  y=a0/2+Σ{ak・cos(ωk・t)+bk・sin(ωk・t)}=A0+ΣAk・cos(ωk・t-θk)\n";
  fout << "    k=1,…,p:周期成分  A0:メサー(定数)  Ak:振幅  ωk=360°・k/T:角周波数\n";
  fout << "    θk:位相  T=" << opt.t0() << ":基本周期  T/k:周期  f=1/T:基本周波数  fk=k/T:周波数\n";
  if (opt.res("statg"))
  {
    int gno = 1;
    for (mapsg_itr gitr = (var.gsmap()).begin(); gitr != (var.gsmap()).end(); ++gitr, ++gno)
    {
      fout << "\n・群" << setw(5) << gno << ":" << id.item(0) << "=" << gitr->first << "\n";
      put_pfun(fout, opt, id, gitr->second, var.gstatcom());
    }
  }
  if (opt.res("statt"))
  {
    fout << "\n・全体\n";
    put_pfun(fout, opt, id, var.gstatall(), var.gstatcom());
  }
}

void pcovaclass::put_pfun(ostream& fout, const option& opt, const idd& id, gstat& gstati, gstat& gstatcom) const
// 周期関数パラメターを出力する
{
  size_t p = id.ni() - 1;
  constexpr int LENline = 78;   // 区切り線の長さ
  putline(fout, LENline);
  for (size_t j = 0; j < p; ++j)
  {
    fout << "t=" << setw(7) << opt.t(j);
    fout << ":";  putmsd(fout, gstati.ystat(j).n(), gstati.ystat(j).mean(), gstati.ystat(j).vr());
  }
  putline(fout, LENline);
  fout << "全体     :";
  putmsd(fout, gstati.ystat(p).n(), gstati.ystat(p).mean(), gstati.ystat(p).vr());
  fout << "\n周期成分      振幅  位相(度)  位相時間            回帰係数    角周波数    周期\n";
  fout << "種類:  k        Ak      θk°  θk/ωk          ak          bk     ωk     T/k\n";
  putline(fout, LENline);
  fout << "群別:  0" << setw(10) << gstati.amp(0);
  fout << "(メサー)\n";
  double mc = gstati.mp();
  constexpr double Deg = 180.0 / M_PI;
  for (int k = 1; k <= gstati.m(); ++k)
  {
    double acr = gstati.acr(k), w = gstati.w(k);
    fout << "群別:" << setw(3) << gstati.pcn(k) << setw(10) << gstati.amp(k);
    fout << setw(10) << acr * Deg << setw(10) << acr / w;
    fout << setw(12) << gstati.a(k) << setw(12) << gstati.b(k);
    fout << setw(8) << w * Deg << setw(8) << gstati.t(k);
    fout << "\n";
    mc -= gstatcom.a(k) * gstati.ma(k) + gstatcom.b(k) * gstati.mb(k);
  }
  putline(fout, LENline);
  fout << "共通:  0" << setw(10) << mc;
  fout << "(メサー)\n";
  for (int k = 1; k <= gstatcom.m(); ++k)
  {
    double acr = gstatcom.acr(k), w = gstatcom.w(k);
    fout << "共通:" <<setw(3) << gstatcom.pcn(k) << setw(10) << gstatcom.amp(k);
    fout << setw(10) << acr * Deg << setw(10) << acr / w;
    fout << setw(12) << gstatcom.a(k) << setw(12) << gstatcom.b(k);
    fout << setw(8) << w * Deg << setw(8) << gstatcom.t(k);
    fout << "\n";
  }
  putline(fout, LENline);
  fout << "群別周期回帰式の重寄与率:R2乗=" << gstati.r2();
  fout << "  重相関係数:R=" << gstati.r() << endl;
}

void pcovaclass::put_pcor(ostream& fout, varclass& var) const
// 周期関数の位相と振幅の一致係数を出力する
{
  if (var.ng() < 2) return;
  constexpr int Nc = 3;         // 一致係数行列の横の個数
  vector<gstat*> gstatpvec(var.ng());
  vecdbl sxxvec(var.ng());
  int i = 0;
  for (mapsg_itr gitr = (var.gsmap()).begin(); gitr != (var.gsmap()).end(); ++gitr, ++i)
  {
    gstatpvec[i] = &(gitr->second);
    sxxvec[i] = pfinp(gitr->second, gitr->second);
  }
  fout << "\n・群別周期回帰式の一致係数";
  int low = 0;
  while (low < var.ng())
  {
    i = low + Nc - 1;
    int up = (i < var.ng() - 1) ? i : var.ng() - 1;
    fout << "\n    ";
    for (i = low; i <= up; ++i)
    {
      fout << "        群" << setw(2) << i + 1;
      fout << "          ";
    }
    fout << "\n";
    fout << "    ";
    for (i = low; i <= up; ++i) fout << "   位相   振幅   全体 ";
    fout << "\n";
    int llen = 4 + (up - low + 1) * 22;
    putline(fout, llen);
    for (i = 0; i < var.ng(); ++i)
    {
      fout << "群" << setw(2) << i + 1;
      for (int j = low; j <= up; ++j)
      {
        if (i == j)
        { fout << setw(8) << "1   " << setw(7) << "1   " << setw(7) << "1   "; }
        else if (sxxvec[i] <= 0.0 || sxxvec[j] <= 0.0)
        {
          int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
          fout.setf(ios::fixed);
          fout << setw(8) << setprecision(3) << 0.0 << setw(7) << 0.0;
          fout << setprecision(prec0);
          fout.unsetf(ios::fixed);
        }
        else
        {
          double sdxij = sqrt(sxxvec[i] * sxxvec[j]);
          double r = (sdxij > 0.0) ? pfinp(*(gstatpvec[i]), *(gstatpvec[j])) / sdxij : 0.0;
          double ra = (sdxij > 0.0) ? 2.0 * sdxij / (sxxvec[i] + sxxvec[j]) : 0.0;
          int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
          fout.setf(ios::fixed);
          fout << setw(8) << setprecision(3) << r << setw(7) << ra << setw(7) << r * ra;
          fout << setprecision(prec0);
          fout.unsetf(ios::fixed);
        }
      }
      fout << "\n";
    }
    putline(fout, llen);
    low = up + 1;
  }
}

void pcovaclass::put_pcovatab0(ostream& fout, const varclass& var) const
// 周期共分散分析表を出力する
{
  string tstr("周期共分散分析表(PERCOVA table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << "\n" << setw(55 + npad) << tstr;
  fout << "\n";
  fout << "要因              平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, var.sa(), var.dfa(), var.sr(), var.dfr());
  tstr = "群差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sa();
  fout << setw(10) << var.dfa() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, var.sbc(), var.dfb(), var.sr(), var.dfr());
  tstr = "共通周期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sbc();
  fout << setw(10) << var.dfb() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.saa(), var.dfa(), var.sr(), var.dfr());
  tstr = "修正群差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.saa();
  fout << setw(10) << var.dfa() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, var.sb0(), var.dfb(), var.sr(), var.dfr());
  tstr = "全体周期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sb0();
  fout << setw(10) << var.dfb() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sd(), var.dfd(), var.sr(), var.dfr());
  tstr = "非平行性";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sd();
  fout << setw(10) << var.dfd() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  tstr = "残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sr();
  fout << setw(10) << var.dfr() << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "全体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.st();
  fout << setw(10) << var.dft();
  fout << "\n";
}

void pcovaclass::put_pcovatab1(ostream& fout, const varclass& var) const
// 一元配置型共分散分析表を出力する
{
  if (var.dfr() > 0)            // 繰り返しの有る一元配置型
  {
    put_pcovatab1rn(fout, var.sa(), var.dfa(), var.sbc(), var.sb0(), var.dfb(), var.saa(), var.sd(), var.dfd(), var.sl(), var.dfl(), var.sp(), var.dfp(), var.sr(), var.dfr(), var.st(), var.dft());
  }
  else                          // 繰り返しの無い一元配置型
  {
    double sl = 0.0, st = var.sp();
    int dfl = 0, dft = var.dfp();
    double sr = var.st() - var.sa() - var.sbc() - var.sd();
    int dfr = var.dft() - var.dfa() - var.dfb() - var.dfd();
    put_pcovatab1rn(fout, var.sa(), var.dfa(), var.sbc(), var.sb0(), var.dfb(), var.saa(), var.sd(), var.dfd(), sl, dfl, var.sp(), var.dfp(), sr, dfr, st, dft);
  }
}

void pcovaclass::put_pcovatab1rn(ostream& fout, double sa, int dfa, double sbc, double sb0, int dfb, double saa, double sd, int dfd, double sl, int dfl, double sp, int dfp, double sr, int dfr, double st, int dft) const
// 繰り返しの有る／無い一元配置型周期共分散分析表を出力する
{
  string tstr("一元配置型周期共分散分析表(PERCOVA1 table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << "\n" << setw(60 + npad) << tstr;
  fout << "\n";
  fout << "要因              平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, sa, dfa, sr, dfr);
  tstr = "群差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sbc, dfb, sr, dfr);
  tstr = "共通周期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sbc;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, saa, dfa, sr, dfr);
  tstr = "修正群差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << saa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sb0, dfb, sr, dfr);
  tstr = "全体周期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sb0;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, sd, dfd, sr, dfr);
  tstr = "非平行性";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sd;
  fout << setw(10) << dfd << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  if (dfl > 0)
  {
    calvfp(v, f, pr, sl, dfl, sr, dfr);
    tstr = "ズレ合計";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
    fout << setw(10) << dfl << setw(14) << v << setw(12) << f << setw(14) << pr;
    fout << pind(pr) << "\n";
    putline(fout, LENline);
    calvfp(v, f, pr, sp, dfp, sr, dfr);
    tstr = "時期";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
    fout << setw(10) << dfp << setw(14) << v << setw(12) << f << setw(14) << pr;
    fout << pind(pr) << "\n";
  }
  tstr = "残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "全体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft;
  fout << "\n";
}

void pcovaclass::put_pcovatab2(ostream& fout, const option& opt, const varclass& var) const
// 二元配置型周期共分散分析表を出力する
{
  string tstr("二元配置型周期共分散分析表(PERCOVA2 table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << "\n" << setw(60 + npad) << tstr;
  fout << "\n";
  fout << "要因              平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  if (opt.pm() == pm3) calvfp(v, f, pr, var.sa(), var.dfa(), var.sr(), var.dfr());
  else calvfp(v, f, pr, var.sa(), var.dfa(), var.ssr(), var.dfsr());
  tstr = "群差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sa();
  fout << setw(10) << var.dfa() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, var.ssr(), var.dfsr(), var.sr(), var.dfr());
  tstr = "個体残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.ssr();
  fout << setw(10) << var.dfsr() << setw(14) << v;
  if (opt.pm() == pm3)
  {
    fout << setw(12) << f << setw(14) << pr;
    fout << pind(pr);
  }
  fout << "\n";
  putline(fout, LENline);
  double vr = calvfp(v, f, pr, var.ss(), var.dfs(), var.sr(), var.dfr());
  tstr = "個体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.ss();
  fout << setw(10) << var.dfs() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, var.sb0(), var.dfb(), var.sr(), var.dfr());
  tstr = "全体周期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sb0();
  fout << setw(10) << var.dfb() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, var.sd(), var.dfd(), var.sr(), var.dfr());
  tstr = "非平行性";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sd();
  fout << setw(10) << var.dfd() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, var.sl(), var.dfl(), var.sr(), var.dfr());
  tstr = "ズレ合計";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sl();
  fout << setw(10) << var.dfl() << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  tstr = "残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sr();
  fout << setw(10) << var.dfr() << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "全体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.st();
  fout << setw(10) << var.dft();
  fout << "\n";
}

void pcovaclass::put_mmcmp(ostream& fout, const option& opt, varclass& var) const
// メサー(=修正群差)の多重比較を出力する
{
  if (var.ssr() <= 0.0) return;  // 出力なし
  int g = var.ng(), p = var.np(), df1 = 1, dc = 1;
  string method = "Tukey";
  switch (opt.mm())
  {
  case mcbon : method = "Bonferroni";  dc = g * (g - 1) / 2;
    break;
  case mcfis : method = "Fisher";
    break;
  case mcsch : method = "Scheffe";  df1 = g - 1;
  }
  fout << "\n・修正群差の" << method << "型多重比較:共変数=全体平均値  ";
  if (opt.mm() == mctuk)
  {
    fout << "群数=" << g << "  自由度=" << var.dfsr() << "\n";
    fout << "群:  修正平均値 - 群:  修正平均値    修正群差         q値   有意確率p値\n";
  }
  else
  {
    fout << "第1自由度=" << df1 << "  第2自由度=" << var.dfsr() << "\n";
    fout << "群:  修正平均値 - 群:  修正平均値    修正群差         F値   有意確率p値\n";
  }
  constexpr int LENline = 75;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = var.ssr() / var.dfsr();
  vec<double> dxv(2 * opt.m());
  double yc = 0.0;
  for (int k = 1; k <= opt.m(); ++k) yc += (var.gstatcom()).a(k) * (var.gstatall()).ma(k) + (var.gstatcom()).b(k) * (var.gstatall()).mb(k);
  mapsg_itr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    gstat* gstatpi = &(gitri->second);
    double ni = (double)(gstatpi->np());
    double bc0i = gstatpi->mp();
    for (int k = 1; k <= opt.m(); ++k) bc0i -= (var.gstatcom()).a(k) * gstatpi->ma(k) + (var.gstatcom()).b(k) * gstatpi->mb(k);
    double yi = bc0i + yc;
    mapsg_itr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      gstat* gstatpj = &(gitrj->second);
      double bc0j = gstatpj->mp();
      for (int k = 1; k <= opt.m(); ++k)
      {
        bc0j -= (var.gstatcom()).a(k) * gstatpj->ma(k) + (var.gstatcom()).b(k) * gstatpj->mb(k);
        int l = (k - 1) * 2;
        dxv.assign(l, gstatpj->ma(k) - gstatpi->ma(k));
        dxv.assign(l + 1, gstatpj->mb(k) - gstatpi->mb(k));
      }
      double yj = bc0j + yc;
      double dy = bc0j - bc0i;
      double vd = (1.0 / ni + 1.0 / (double)(gstatpj->np()) + (var.gstatall()).vxr(dxv)) * vr;
      double f = dy * dy / vd / (double)df1;
      ldbl pr = calmcpr(opt.mm(), f, g, df1, var.dfsr(), dc);
      fout << setw(2) << j + 1 << ":" << setw(12) << yj << " -";
      fout << setw(3) << i + 1 << ":" << setw(12) << yi << setw(12) << dy;
      fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

void pcovaclass::put_mmci(ostream& fout, const option& opt, varclass& var) const
// メサー(=修正群差)の信頼区間出力を出力する
{
  if (var.ssr() <= 0.0) return;  // 出力なし
  int g = var.ng(), p = var.np(), df1 = 1, dc = 1;
  string method = "Tukey";
  switch (opt.mm())
  {
  case mcbon : method = "Bonferroni";  dc = g * (g - 1) / 2;
    break;
  case mcfis : method = "Fisher";
    break;
  case mcsch : method = "Scheffe";  df1 = g - 1;
  }
  fout << "\n・修正群差の" << method << "型" << opt.cc() * 100.0 << "%同時信頼区間(simultaneous confidence interval)\n";
  fout << "群 - 群    修正群差    絶対精度        下限        上限\n";
  const int LENline = 55;       // 区切り線の長さ
  putline(fout, LENline);
  double vr = var.ssr() / var.dfsr();
  vec<double> dxv(2 * opt.m());
  mapsg_itr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    gstat* gstatpi = &(gitri->second);
    double ni = (double)(gstatpi->np());
    double bc0i = gstatpi->mp();
    for (int k = 1; k <= opt.m(); ++k) bc0i -= (var.gstatcom()).a(k) * gstatpi->ma(k) + (var.gstatcom()).b(k) * gstatpi->mb(k);
    mapsg_itr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      gstat* gstatpj = &(gitrj->second);
      double bc0j = gstatpj->mp();
      for (int k = 1; k <= opt.m(); ++k)
      {
        bc0j -= (var.gstatcom()).a(k) * gstatpj->ma(k) + (var.gstatcom()).b(k) * gstatpj->mb(k);
        int l = (k - 1) * 2;
        dxv.assign(l, gstatpj->ma(k) - gstatpi->ma(k));
        dxv.assign(l + 1, gstatpj->mb(k) - gstatpi->mb(k));
      }
      double dy = bc0j - bc0i;
      double vd = (1.0 / ni + 1.0 / (double)(gstatpj->np()) + (var.gstatall()).vxr(dxv)) * vr;
      double ci = calmcci(opt.mm(), opt.cc(), g, df1, var.dfsr(), dc, vd);
      fout << setw(2) << j + 1 << " -" << setw(3) << i + 1 << setw(12) << dy;
      fout << setw(12) << ci << setw(12) << dy - ci << setw(12) << dy + ci;
      fout << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

void pcovaclass::put_mdcmp(ostream& fout, const option& opt, varclass& var) const
// 非平行性の多重比較を出力する
{
  if (var.sr() <= 0.0) return;  // 出力なし
  int g = var.ng(), p = var.np(), dfb = 2 * opt.m();
  if (2 * opt.pcmax() >= p) --dfb;
  int df1 = dfb, dc = dc = g * (g - 1) / 2;
  string method = "Bonferroni";
  switch (opt.md())
  {
  case mcfis : method = "Fisher";  dc = 1;
    break;
  case mcsch : method = "Scheffe";  df1 = (g - 1) * dfb;  dc = 1;
  }
  fout << "\n・非平行性の" << method << "型多重比較:第1自由度=" << df1 << "  第2自由度=" << var.dfr() << "\n";
  fout << "群 - 群  回帰係数差平方和         F値   有意確率p値\n";
  constexpr int LENline = 55;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = var.sr() / var.dfr();
  vec<double> dbv(2 * opt.m());
  mapsg_itr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    gstat* gstatpi = &(gitri->second);
    mapsg_itr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      gstat* gstatpj = &(gitrj->second);
      gstat gstatc(p);
      gstatc += *(gstatpi);  gstatc += *(gstatpj);
      gstatc.calpfun(opt, false);
      for (int k = 1; k <= opt.m(); ++k)
      {
        int l = (k - 1) * 2;
        dbv.assign(l, gstatpj->a(k) - gstatpi->a(k));
        dbv.assign(l + 1, gstatpj->b(k) - gstatpi->b(k));
      }
      double sdb = gstatc.vxr(dbv);
      double f = sdb / vr / (double)df1;
      double pr = calmcpr(opt.md(), f, g, df1, var.dfr(), dc);
      fout << setw(2) << j + 1 << " -" << setw(3) << i + 1;
      fout << setw(18) << sdb << setw(12) << f << setw(14) << pr;
      fout << pind(pr) << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

int pcovaclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー");
  return nc;
}

void pcovaclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (opt.res("res") == false) return;
  put_id(cout, id);
  if (id.nc() < 1)              // データなし
  {
    cout << "\n" << endl;
    return;
  }
  varclass var(id.ni());        // 変数オブジェクト
  if (var.cal(opt, datalst) < 1) return; // 群なし
  if (opt.res("statg") || opt.res("statt")) put_mreg(cout, opt, id, var);
  if (opt.res("r")) put_pcor(cout, var);
  if (opt.res("atab"))
  {
    switch (opt.pm())
    {
    case pm0 : put_pcovatab0(cout, var);
      break;
    case pm1 : put_pcovatab1(cout, var);
      break;
    default : put_pcovatab2(cout, opt, var);
      if (opt.pm() == pm3) var.pm3(); // 個体残差に残差を代入する
    }
  }
  if (opt.res("mm"))
  {
    put_mmcmp(cout, opt, var);
    put_mmci(cout, opt, var);
  }
  if (opt.res("md")) put_mdcmp(cout, opt, var);
  cout << "\n" << endl;
}

int main(int argc, char *argv[]) // メインルーチン
{
  pcovaclass pcova;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) pcova.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { pcova.get(argv[i], opt, id, datalst);  fok = false; }
    }
    pcova.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
