// preg.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2023-12-15 11:28:47 tac>
//
// 周期回帰分析を行う。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W preg preg.C -I/usr/local/include
//               -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H,mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "preg  Version 7.42           Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:preg  [<オプション> …]  [<データファイル> …]",
  "機能:全項目を対象とした周期回帰分析を行います。",
  "<オプション>",
  "  -dt=<時期間隔>:時間項目の時間間隔を指定する。(オプション省略時:-dt=T/項目数)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込みます。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示します。(-h:-h=man)",
  "  -k=<番号>[,<番号>,…]:周期回帰式に組み込む周期成分番号を指定します。",
  "  -ngp={mse|prc}[,…]:Ngraph用出力データを指定します。",
  "    mse:平均値±SE  prc:周期回帰曲線  オプション省略時:全て指定",
  "  -ngpname[=<出力ファイル名>]:Ngraph用データの出力ファイル名を指定します。",
  "    オプション省略時:Ngraph用データ出力せず  -ngpname:-ngpname=preg_ngp.dat",
  "  -pm={0|1|2}:周期解析の手法を指定します。",
  "  -res=[-][{def|all|cc[<信頼係数>]|dw|h[{a|d|n}][{m|<基準値>}]",
  "        |lv[{m|<基準値>}]|pk|reg|stat}][,…]:解析結果の出力項目を指定します。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    cc[<信頼係数>]:メサーの信頼区間(0≦信頼係数＜1)  dw:ダービン・ワトソン比",
  "    h[{a|d|n}][{m|<基準値>}]:HBI(a:全日 d:昼間 n:夜間 m:メサー)",
  "    lv[{m|<基準値>}]:基準値時間(m:メサー)  pk:周期関数の特徴値",
  "    reg:周期回帰分析結果  stat:基礎統計量",
  "  -T=<基本周期>:基本周期(1周期の時間)を指定します。",
  "  -t=<時間1>,<時間2>[,<時間3>,…]:時期項目の時間を指定します。",
  "  -t{d|n}=<開始時間>:昼間・夜間の開始時間を指定します。"
};

const string Title = "=== 周期回帰分析(periodic regression analysis) ==="; // 表題
const string NGPNAMEdef = "preg_ngp.dat"; // 省略時Ngraphファイル名
constexpr double T0def = 24.0;  // 省略時基本周期
constexpr double TDdef = 7.0;   // 省略時昼の開始時間
constexpr double TNdef = 23.0;  // 省略時夜の開始時間
constexpr uint Dev = 720;       // 関数サーチ時の分解数

using vecstatv = vector<statv<double>>; // 基礎統計量ベクトルの短名定義

enum pmtype                     // 周期解析の種類
{
  pmnon,                        // なし
  pm0,                          // 平均値を用いる単純法
  pm1,                          // 一元配置法
  pm2,                          // 二元配置法
  pmetc                         // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  double vdt;                   // 時期間隔
  double vha, vhd, vhn;         // 全日、昼間、夜間HBI計算用基準値
  double vlv;                   // 基準値時間出力用基準値
  string vngpname;              // Ngraphファイル基本名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr pclst;                   // 周期成分番号リスト
  vecint vpcvec;                // 周期成分番号配列
  int vpcmax;                   // 周期成分番号最大値
  pmtype vpm;                   // 周期解析の種類
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  lstr tlst;                    // 項目時間リスト
  vecdbl vtvec;                 // 項目時間ベクトル
  double vt0;                   // 基本周期
  double vtd, vtn;              // 昼間、夜間開始時間
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double dt() const             // 時期間隔を返す
  { return vdt; }
  double ha() const             // 全日HBI計算用基準値を返す
  { return vha; }
  double hd() const             // 昼間HBI計算用基準値を返す
  { return vhd; }
  double hn() const             // 夜間HBI計算用基準値を返す
  { return vhn; }
  double lv() const             // 基準値時間出力用基準値を返す
  { return vlv; }
  int m() const                 // 周期成分数を返す
  { return vpcvec.size(); }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  int pcn(size_t k) const       // 周期成分番号を返す
  { return (k < vpcvec.size()) ? vpcvec[k] : 0; }
  const vecint& pcvec() const   // 周期成分番号ベクトルを返す
  { return vpcvec; }
  int pcmax() const             // 周期成分番号最大値を返す
  { return vpcmax; }
  pmtype pm() const             // 周期解析の種類を返す
  { return vpm; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t(size_t idx) const    // 項目時間を返す
  { return (idx < vtvec.size()) ? vtvec[idx] : vt0; }
  double t0() const             // 基本周期を返す
  { return vt0; }
  double td() const             // 昼間開始時間を返す
  { return vtd; }
  double tn() const             // 夜間開始時間を返す
  { return vtn; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdt(0.0), vha(BLANKf), vhd(BLANKf), vhn(BLANKf), vlv(BLANKf), vngpname(), vngplst(), vngpmap(), pclst(), vpcvec(), vpcmax(-1), vpm(pm2), vreslst(), vresmap(), tlst(), vtvec(), vt0(T0def), vtd(TDdef), vtn(TNdef)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dt")           // 時期間隔の指定
  {
    vdt = strtodf(opt[1]);
    if (vdt <= 0.0 || vdt == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "k")       // 周期成分番号の指定
  {
    if (strsplit(pclst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pm")      // 周期解析の手法指定
  {
    if (opt[1].empty() || opt[1] == "2") vpm = pm2;
    else if (opt[1] == "0") vpm = pm0;
    else if (opt[1] == "1") vpm = pm1;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "T")       // 基本周期の指定
  {
    vt0 = strtodf(opt[1]);
    if (vt0 == BLANKf || vt0 <= 0.0) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "t")       // 項目時間の指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "td")      // 昼間開始時間の指定
  {
    vtd = opt[1].empty() ? TDdef : strtodf(opt[1]);
    if (vtd == BLANKf || vtd < 0.0 || 24.0 <= vtd) err = badopt(vpname, optstr, USAGE, 0.0, 24.0);
  }
  else if (opt[0] == "tn")      // 夜間開始時間の指定
  {
    vtn = opt[1].empty() ? TNdef : strtodf(opt[1]);
    if (vtn == BLANKf || vtn <= vtd || 24.0 <= vtn) err = badopt(vpname, optstr, USAGE, vtd + 0.01, 24.0);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni < 2) return disperr(vpname, "時期項目が2つ以上必要です！");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  int q = ni / 2;                // 周期成分最大数
  vtvec.resize(ni);
  double dt = (vdt > 0.0) ? vdt : vt0 / (double)ni;
  for (size_t j = 0; j < ni; ++j) vtvec[j] = (double)j * dt;
  if (settvec(vtvec, tlst, ni) < 0) return disperr(vpname, "項目時間が間違っています！");
  if (pclst.empty())            // 周期成分指定がない時
  {
    vpcvec.resize(2);
    vpcvec[0] = 1;  vpcvec[1] = 2;  vpcmax = 2;
  }
  else if ((vpcmax = setpcvec(vpcvec, pclst, q)) < 0)
  { return disperr(vpname, "周期成分指定が間違っています！"); }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["mse"] = vngpmap["prc"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "-ngpnameオプションの指定が必要！");
    vngpmap["mse"] = vngpmap["prc"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "-ngp=" + ngp + "オプション指定が不適切！"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = true;
  vresmap["dw"] = vresmap["ha"] = vresmap["hd"] = vresmap["hn"] = vresmap["lv"] = vresmap["pk"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "-res=" + res + "オプション指定が不適切！"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = vresmap["dw"] = vresmap["ha"] = vresmap["hd"] = vresmap["hn"] = vresmap["lv"] = vresmap["pk"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        vresmap["cc"] = true;
      }
      else if (res == "h")      // -reg=hの時
      { vresmap["ha"] = vresmap["hd"] = vresmap["hn"] = true; }
      else if (res == "hm")      // -reg=hmの時
      {
        vresmap["ha"] = vresmap["hd"] = vresmap["hn"] = true;
        vha = vhd = vhn = BLANKf;
      }
      else if (res == "ha" || res == "hd" || res == "hn")
      { vresmap[res] = true; }
      else if (res[0] == 'h' && res[1] == 'a' && res.size() > 2)
      {
        vha = strtodf(res.substr(2)); // 基準値
        if (vha == BLANKf) return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        vresmap["ha"] = true;
      }
      else if (res[0] == 'h' && res[1] == 'd' && res.size() > 2)
      {
        vhd = strtodf(res.substr(2)); // 基準値
        if (vhd == BLANKf) return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        vresmap["hd"] = true;
      }
      else if (res[0] == 'h' && res[1] == 'n' && res.size() > 2)
      {
        vhn = strtodf(res.substr(2)); // 基準値
        if (vhn == BLANKf) return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        vresmap["hn"] = true;
      }
      else if (res[0] == 'h' && res.size() > 1) // -res=h<基準値>の時
      {
        vha = strtodf(res.substr(1)); // 基準値
        if (vha == BLANKf) return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        vhd = vhn = vha;
        vresmap["ha"] = vresmap["hd"] = vresmap["hn"] = true;
      }
      else if (res == "lvm")
      { vresmap["lv"] = true;  vlv = BLANKf; }
      else if (res[0] == 'l' && res[1] == 'v' && res.size() > 2)
      {
        vlv = strtodf(res.substr(2)); // 基準値
        if (vlv == BLANKf) return disperr(vpname, "-res=" + res + "オプション指定が不適切！");
        vresmap["lv"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "-res=" + res + "オプション指定が不適切！"); }
      else
      { vresmap[res] = true; }
    }
  }
  if (vresmap["ha"] == false && (vresmap["hd"] || vresmap["hn"])) vresmap["ha"] = true;
  if (vresmap["hd"] == false) vresmap["hd"] = vresmap["ha"];
  if (vresmap["hn"] == false) vresmap["hn"] = vresmap["ha"];
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  int vnp;                      // 時期項目数
  vecstatv ystatvec;            // 時期項目の基本統計量ベクトル
  int vnb;                      // 観察個体数
  double vsb;                   // 観察個体平方和
  mat<double> smat;             // 周期回帰計算用積和行列
  pfun<double> vpf;             // 周期関数オブジェクト
  double vst;                   // 全体平方和
  int vdft;                     // 全体自由度
  double vsa;                   // 群差平方和
  int vdfa;                     // 群差自由度
  int vdfb;                     // 回帰自由度
  double vsp;                   // 時期平方和
  int vdfp;                     // 時期自由度
  double vsl;                   // ズレ平方和
  int vdfl;                     // ズレ自由度
  double vsr;                   // 残差平方和
  int vdfr;                     // 残差自由度
  int cal_preg(const option& opt);
  // 基礎統計・周期回帰式の各種パラメターを計算する
public:
  varclass(size_t ni) : vnp(ni), ystatvec(ni + 1), vnb(0), vsb(0.0), smat(), vpf(), vst(0.0), vdft(0), vsa(0.0), vdfa(0), vdfb(0), vsp(0.0), vdfp(0), vsl(0.0), vdfl(0), vsr(0.0), vdfr(0) // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int np() const                // 時期項目数を返す
  { return vnp; }
  int nb() const                // 観察個体数を返す
  { return vnb; }
  int n(size_t idx) const       // 時期項目の例数を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].n() : 0; }
  double mean(size_t idx) const // 時期項目の平均を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].mean() : 0.0; }
  double vr(size_t idx) const   // 時期項目の不偏分散を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].vr() : 0.0; }
  double se(size_t idx) const   // 時期項目の標準誤差を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].se() : 0.0; }
  double sb() const             // 観察個体平方和を返す
  { return vsb; }
  double ss(size_t iidx, size_t jidx) const // 積和行列の要素を返す
  { return (iidx < smat.r() && jidx < smat.c()) ? smat(iidx, jidx) : 0.0; }
  pfun<double>& pf()            // 周期関数オブジェクトを返す
  { return vpf; }
  double st() const             // 全体平方和を返す
  { return vst; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
  double sa() const             // 群差平方和を返す
  { return vsa; }
  int dfa() const               // 群差自由度を返す
  { return vdfa; }
  int dfb() const               // 回帰自由度を返す
  { return vdfb; }
  double sp() const             // 時期平方和を返す
  { return vsp; }
  int dfp() const               // 時期自由度を返す
  { return vdfp; }
  double sl() const             // ズレ平方和を返す
  { return vsl; }
  int dfl() const               // ズレ自由度を返す
  { return vdfl; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  int cal(const option& opt, ldata& datalst); // データを集計して統計量を求める
  double mhbi(const option& opt, double ts, double te, double level) const;
  // 実測平均によるHBIを返す
  double pfhbi(double ts, double te, double level);
  // 周期関数によるHBIを返す
  int sltime(vecdbl& tlvec, double ts, double te, double level) const;
  // 周期回帰式基準値時間をサーチする
};

int varclass::cal_preg(const option& opt)
// 基礎統計・周期回帰式の各種パラメターを計算する
{
  int p = vnp, dim = 2 * opt.m() + 2, np = 0;
  int yidx = dim - 1;           // yのインデックス
  vecdbl xyvec(dim);
  double w = 2.0 * M_PI / opt.t0();
  xyvec[0] = 1.0;  vsa = vst = 0.0;
  for (int j = 0; j < p; ++j)
  {
    int n = ystatvec[j].n();
    if (n > 0)
    {
      double me = ystatvec[j].mean();
      xyvec[yidx] = me;  ystatvec[p] += ystatvec[j];
      for (int k = 0; k < opt.m(); ++k)
      {
        int i = 2 * (k + 1);
        double tk = w * opt.pcn(k) * opt.t(j);
        xyvec[i - 1] = cos(tk);
        xyvec[i] = (2 * opt.pcn(k) < p) ? sin(tk) : 0.0;
      }
      int r = (opt.pm() == pm0) ? 1 : n;
      for (int i = 0; i < dim; ++i)
      {
        double x = r * xyvec[i];
        for (int k = i; k < dim; ++k) smat.add(i, k, x * xyvec[k]);
      }
      ++np;  vsa += n * me * me;
    }
  }
  if (np < yidx) return disperr(opt.pname(), "時期項目が少なすぎます！");
  for (int i = 0; i < yidx; ++i)
  {
    for (int j = i + 1; j < dim; ++j) smat.assign(j, i, smat(i, j));
  }
  smat.sweep(0);
  if (opt.pm() != pm0) smat.assign(yidx, yidx, ystatvec[p].ss());
  vst = smat(yidx, yidx);
  vdfa = np - 1;
  vdft = (opt.pm() == pm0) ? vdfa : ystatvec[p].n() - 1;
  for (int j = 1; j < yidx; ++j) smat.sweep(j);
  double cf = ystatvec[p].n() * ystatvec[p].mean() * ystatvec[p].mean();
  vsa -= cf;  vsb = vsb / (double)p - cf;  vsp = vst - smat(yidx, yidx);
  vsl = vsa - vsp;
  double eps = DBL_EPSILON * fmax(5.0, vst);
  if (vsa < eps) vsa = 0.0;
  if (vsb < eps) vsb = 0.0;
  if (vsp < eps) vsp = 0.0;
  if (vsl < eps) vsl = 0.0;
  vdfp = 2 * opt.m();
  if (2 * opt.pcmax() >= p) --vdfp;
  vdfl = vdfa - vdfp;  vdfb = vnb - 1;  vdfr = vdft - vdfp;
  if (vdfl < 1) vsl = 0.0;
  vsr = vst - vsp;
  if (opt.pm() == pm1) { vsr = vst - vsa;  vdfr = vdft - vdfa; }
  else if (opt.pm() == pm2) { vsr = vst - vsa - vsb;  vdfr = vdft - vdfa - vdfb; }
  if (vdfr < 1 || vsr < eps) vsr = 0.0;
  return np;                    // 有効な時期項目数を返す
}

int varclass::cal(const option& opt, ldata& datalst)
// データを集計して統計量を求める
{
  vecdbl fdatavec(vnp);
  for (size_t j = 0; j < ystatvec.size(); ++j) ystatvec[j].clear();
  vnb = 0;  vsb = 0.0;
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    bool blank = false;
    for (int j = 0; j < vnp; ++j)
    {
      fdatavec[j] = ditr->fdata(j);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (blank == false || opt.pm() != pm2)
    {
      double sum = 0.0;
      for (int j = 0; j < vnp; ++j)
      {
        if (fdatavec[j] != BLANKf)
        { ystatvec[j].sigma(fdatavec[j]);  sum += fdatavec[j]; }
      }
      ++vnb;  vsb += sum * sum;
    }
  }
  int dim = 2 * opt.m() + 2;
  smat.resize(dim, dim);        // 周期回帰計算用積和行列
  int np = cal_preg(opt);
  vecdbl abvec(dim - 1);
  for (int j = 0; j < dim - 1; ++j) abvec[j] = smat(j, dim - 1);
  pfun<double> pf(opt.t0(), opt.pcvec(), abvec);
  vpf = pf;
  return np;                    // 有効な時期項目数を返す
}

double varclass::mhbi(const option& opt, double ts, double te, double level) const
// 実測平均によるHBIを返す
{
  int i = 0, p = vnp;
  while (ts >= opt.t(i) && i < p) ++i;
  double to = opt.t(p - 1), ti = 2 * opt.t(p - 1) - opt.t(p - 2);
  double mo = ystatvec[p - 1].mean(), m = ystatvec[0].mean();
  if (i < 1)
  {
    to = 2 * opt.t(0) - opt.t(1);  ti = opt.t(0);
    mo = ystatvec[p - 1].mean();  m = ystatvec[0].mean();
  }
  else if (i < p)
  {
    to = opt.t(i - 1);  ti = opt.t(i);
    mo = ystatvec[i - 1].mean();  m = ystatvec[i].mean();
  }
  double a = mo + (m - mo) * (ts - to) / (ti - to) - level;
  to = ts;
  double hbi = 0.0;
  int j = i;
  for (j = i; j < p && opt.t(j) <= te; ++j)
  {
    if (ystatvec[j].n() > 0)
    {
      double dt = opt.t(j) - to, b = ystatvec[j].mean() - level;
      if (a > 0.0 && b > 0.0) hbi += (a + b) * dt;
      else if (a > 0.0) hbi += a * a / (a - b) * dt;
      else if (b > 0.0) hbi += b * b / (b - a) * dt;
      a = b;  to = opt.t(j);
    }
  }
  if (j < p)
  {
    to = opt.t(j - 1);  ti = opt.t(j);
    mo = ystatvec[j - 1].mean();  m = ystatvec[j].mean();
  }
  else
  {
    to = opt.t(p - 1);  ti = 2 * opt.t(p - 1) - opt.t(p - 2);
    mo = ystatvec[p - 1].mean();  m = ystatvec[0].mean();
  }
  double b = mo + (m - mo) * (te - to) / (ti - to) - level, dt = te - to;
  if (a > 0.0 && b > 0.0) hbi += (a + b) * dt;
  else if (a > 0.0) hbi += a * a / (a - b) * dt;
  else if (b > 0.0) hbi += b * b / (b - a) * dt;
  return hbi * 0.5;
}

double varclass::pfhbi(double ts, double te, double level)
// 周期関数によるHBIを返す
{
  vecdbl tlvec(2 * vpf.pcmax());
  int nl = sltime(tlvec, ts, te, level);
  double to = ts, hbi = 0.0;
  for (int i = 0; i < nl; ++i)
  {
    double t = tlvec[i];
    double s = vpf.integ(to, t) - level * (t - to);
    if (s > 0.0) hbi += s;
    to = t;
  }
  if (vpf.fun(te) > level) hbi += vpf.integ(to, te) - level * (te - to);
  return hbi;
}

int varclass::sltime(vecdbl& tlvec, double ts, double te, double level) const
// 周期回帰式基準値時間をサーチする
{
  int nl = 0;
  double to = ts, dt = vpf.t(0) / (double)Dev, yo = vpf.fun(ts);
  double t = ts + dt;
  for (uint i = 1; i <= Dev; ++i, t += dt)
  {
    double y = vpf.fun(t);
    if (yo <= level && level < y)
    {
      tlvec[nl] = (level - yo < y - level) ? to : t;
      ++nl;
    }
    else if (yo >= level && level > y)
    {
      tlvec[nl] = (yo - level < level - y) ? to : t;
      ++nl;
    }
    yo = y;  to = t;
  }
  return nl;                    // 基準値時間数を返す
}

class pregclass                 // pregメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  // 基礎統計・周期回帰式の各種パラメターを計算する
  void put_stat(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 時期別基礎統計量を出力する
  void put_pstat(ostream& fout, option& opt, const idd& id, varclass& var) const;
  // 集計結果を出力する
  void put_pregfun(ostream& fout, const option& opt, const pfun<double>& pf, double st, double sa, double sp) const;
  // 周期関数パラメターを出力する
  void put_pval(ostream& fout, varclass& var) const;
  // 周期回帰式極値と極値時間を出力する
  void put_cval(ostream& fout, varclass& var) const;
  // 周期回帰式変曲点と変曲点時間を出力する
  void put_lval(ostream& fout, const option& opt, varclass& var) const;
  // 周期回帰式基準値時間を出力する
  void put_hbi(ostream& fout, const option& opt, varclass& var) const;
  // HBIを出力する
  void put_dw(ostream& fout, const option& opt, const idd& id, varclass& var) const;
  // ダービン・ワトソン比を出力する
  void put_pregtab1(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sr, int dfr, double st, int dft) const;
  // 一元配置型周期回帰分散分析表を出力する
  void put_pregtab2(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sb, int dfb, double sr, int dfr, double st, int dft) const;
  // 二元配置型周期回帰分散分析表を出力する
  void plot_ngp(option& opt, idd& id, varclass& var) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpmse(ofstream& fout, const option& opt, const idd& id, const varclass& var) const;
  // Ngraph用時期別平均±SEを出力する
  void plot_ngpprc(ofstream& fout, const pfun<double>& pf) const;
  // Ngraph用周期回帰曲線を出力する
public:
  pregclass() : fname()         // コンストラクタ
  {}
  ~pregclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void pregclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "時期項目  1:" << id.item(0) << "\n";
  fout << "  :\n";
  fout << "時期項目" << setw(3) << id.ni();
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void pregclass::put_stat(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 時期別基礎統計量を出力する
{
  const int LENline = 77;
  size_t p = id.ni();
  fout << "\n時間 t   :基礎統計量\n";
  putline(fout, LENline);
  for (size_t j = 0; j < p; ++j)
  {
    int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
    fout.setf(ios::fixed);
    fout << setw(9) << setprecision(2) << opt.t(j);
    fout.unsetf(ios::fixed);
    fout << ":" << setprecision(prec0);
    putmsd(fout, var.n(j), var.mean(j), var.vr(j));
  }
  putline(fout, LENline);
  fout << "全体     :";  putmsd(fout, var.n(p), var.mean(p), var.vr(p));
}

void pregclass::put_pstat(ostream& fout, option& opt, const idd& id, varclass& var) const
// 集計結果を出力する
{
  put_pregfun(fout, opt, var.pf(), var.st(), var.sa(), var.sp());
  double vr = (var.dfr() > 0) ? var.sr() / (double)(var.dfr()) * var.ss(0, 0) : 0.0;
  if (opt.res("cc"))
  {
    double ci = calmcci(mcfis, opt.cc(), 1, var.dfp(), var.dfr(), 1, vr);
    fout << "\n・メサーの" << opt.cc() * 100.0 << "%信頼区間=";
    fout << (var.pf()).amp(0) << "±" << ci << "(" << (var.pf()).amp(0) - ci << "-" << (var.pf()).amp(0) + ci << ")\n";
  }
  if (opt.res("pk"))
  {
    put_pval(fout, var);
    put_cval(fout, var);
  }
  if (opt.res("lv")) put_lval(fout, opt, var);
  if (opt.res("ha")) put_hbi(fout, opt, var);
  if (opt.res("dw") && var.sr() > 0.0) put_dw(fout, opt, id, var);
  if (var.st() > 0.0)
  {
    fout << "\n";
    switch (opt.pm())
    {
    case pm0 : putatab(fout, "周期回帰", var.sp(), var.dfp(), var.sr(), var.dfr(), var.st(), var.dft());
      break;
    case pm1 : put_pregtab1(fout, var.sp(), var.dfp(), var.sl(), var.dfl(), var.sa(), var.dfa(), var.sr(), var.dfr(), var.st(), var.dft());
      break;
    default : put_pregtab2(fout, var.sp(), var.dfp(), var.sl(), var.dfl(), var.sa(), var.dfa(), var.sb(), var.dfb(), var.sr(), var.dfr(), var.st(), var.dft());
    }
  }
}

void pregclass::put_pregfun(ostream& fout, const option& opt, const pfun<double>& pf, double st, double sa, double sp) const
// 周期関数パラメターを出力する
{
  fout << "\n・周期回帰式(periodic regression equation)\n";
  fout << "  y=a0/2+Σ{ak・cos(ωk・t)+bk・sin(ωk・t)}=A0+ΣAk・cos(ωk・t-θk)\n";
  fout << "    k=1,…,p:周期成分  A0:メサー(定数)  Ak:振幅  ωk=360°・k/T:角周波数\n";
  fout << "    θk:位相  T=" << pf.t(0) << ":基本周期  T/k:周期  f=1/T:基本周波数  fk=k/T:周波数\n";
  fout << "\n周期成分      振幅  位相(度)  位相時間            回帰係数    角周波数    周期\n";
  fout << "   k            Ak      θk°  θk/ωk          ak          bk     ωk     T/k\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  fout << "   0  : " << setw(10) << pf.amp(0);
  fout << "(メサー)\n";
  double deg = 180.0 / M_PI;
  for (size_t k = 1; k <= pf.m(); ++k)
  {
    fout << setw(4) << pf.pcn(k) << "  : " << setw(10) << pf.amp(k);
    fout << setw(10) << pf.acr(k) * deg << setw(10) << pf.acr(k) / pf.w(k);
    fout << setw(12) << pf.a(k) << setw(12) << pf.b(k);
    fout << setw(8) << pf.w(k) * deg << setw(8) << pf.t(k);
    fout << "\n";
  }
  putline(fout, LENline);
  double r2 = (st > 0.0) ? sp / st : 0.0;
  if (opt.pm() == pm0)
  {
    fout << "周期回帰式の重寄与率:R2乗=" << r2 << "  重相関係数:R=" << sqrt(r2) << endl;
  }
  else
  {
    r2 = (sa > 0.0) ? sp / sa : 0.0;
    fout << "平均値に対する周期回帰式の重寄与率:R2乗=" << r2 << "  重相関係数:R=" << sqrt(r2) << "\n";
    r2 = (st > 0.0) ? sp / st : 0.0;
    fout << "データに対する周期回帰式の重寄与率:R2乗=" << r2 << "  重相関係数:R=" << sqrt(r2) << endl;
  }
}

void pregclass::put_pval(ostream& fout, varclass& var) const
// 周期回帰式極値と極値時間を出力する
{
  vecdbl tpvec((var.pf()).pcmax()), tnvec((var.pf()).pcmax());
  int npk = 0, nnd = 0;
  double to = 0.0, dt = (var.pf()).t(0) / (double)Dev, dyo = (var.pf()).dfun(0.0);
  double t = dt;
  for (uint i = 1; i <= Dev; ++i, t += dt)
  {
    double dy = (var.pf()).dfun(t);
    if (dyo <= 0.0 && 0.0 < dy) // 極小
    {
      tnvec[nnd] = (-dyo < dy) ? to : t;
      ++nnd;
    }
    else if (dyo >= 0.0 && 0.0 > dy) // 極大
    {
      tpvec[npk] = (dyo < -dy) ? to : t;
      ++npk;
    }
    dyo = dy;  to = t;
  }
  int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
  fout.precision(5);
  fout << "\n・極小値:";
  for (int i = 0; i < nnd; ++i)
  {
    if (i > 0 && i % 3 == 0) fout << "\n         ";
    fout << "(t=" << left << setw(8) << tnvec[i];
    fout << "y=" << setw(8) << (var.pf()).fun(tnvec[i]) << right << ") ";
  }
  fout << "\n  極大値:";
  for (int i = 0; i < npk; ++i)
  {
    if (i > 0 && i % 3 == 0) fout << "\n         ";
    fout << "(t=" << left << setw(8) << tpvec[i];
    fout << "y=" << setw(8) << (var.pf()).fun(tpvec[i]) << right << ") ";
  }
  fout << endl;
  fout.precision(prec0);        // 小数点以下の精度を元に戻す
}

void pregclass::put_cval(ostream& fout, varclass& var) const
// 周期回帰式変曲点と変曲点時間を出力する
{
  vecdbl tcvec(2 * (var.pf()).pcmax());
  int nc = 0;
  double to = 0.0, dt = (var.pf()).t(0) / (double)Dev, d2yo = (var.pf()).d2fun(0.0);
  double t = dt;
  for (uint i = 1; i <= Dev; ++i, t += dt)
  {
    double d2y = (var.pf()).d2fun(t);
    if (d2yo <= 0.0 && 0.0 < d2y) // 勾配が正
    {
      tcvec[nc] = (-d2yo < d2y) ? to : t;
      ++nc;
    }
    else if (d2yo >= 0.0 && 0.0 > d2y) // 勾配が負
    {
      tcvec[nc] = (d2yo < -d2y) ? to : t;
      ++nc;
    }
    d2yo = d2y;  to = t;
  }
  int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
  fout.precision(5);
  fout << "  変曲点:";
  for (int i = 0; i < nc; ++i)
  {
    if (i > 0 && i % 2 == 0) fout << "\n         ";
    fout << "(t=" << left << setw(8) << tcvec[i];
    fout << "y=" << setw(8) << (var.pf()).fun(tcvec[i]);
    fout << "g=" << setw(8) << (var.pf()).dfun(tcvec[i]) << right << ") ";
  }
  fout << endl;
  fout.precision(prec0);        // 小数点以下の精度を元に戻す
}

void pregclass::put_lval(ostream& fout, const option& opt, varclass& var) const
// 周期回帰式基準値時間を出力する
{
  double level = (opt.lv() == BLANKf) ? (var.pf()).amp(0) : opt.lv();
  vecdbl tlvec(2 * (var.pf()).pcmax());
  int nl = var.sltime(tlvec, 0.0, (var.pf()).t(0), level);
  fout << "\n・y=" << left << setw(8) << level << right;
  fout << "となるt=";
  for (int i = 0; i < nl; ++i)
  {
    if (i > 0 && i % 7 == 0) fout << "\n" << setw(20) << "";
    fout << setw(8) << tlvec[i];
  }
  fout << endl;
}

void pregclass::put_hbi(ostream& fout, const option& opt, varclass& var) const
// HBIを出力する
{
  double ha = (opt.ha() == BLANKf) ? (var.pf()).amp(0) : opt.ha();
  double hd = (opt.hd() == BLANKf) ? (var.pf()).amp(0) : opt.hd();
  double hn = (opt.hn() == BLANKf) ? (var.pf()).amp(0) : opt.hn();
  fout << "\n・Hyper Baric Index 基準値:全日=" << ha;
  fout << " 昼間(" << opt.td() << "-" << opt.tn() - 1.0 << ")=" << hd;
  fout << " 夜間(" << opt.tn() << "-" << opt.td() - 1.0 << ")=" << hn << "\n";
  double hbia = var.mhbi(opt, 0.0, (var.pf()).t(0), ha);
  double hbid = var.mhbi(opt, opt.td(), opt.tn(), hd);
  double hbin = var.mhbi(opt, 0.0, (var.pf()).t(0), hn) - var.mhbi(opt, opt.td(), opt.tn(), hn);
  double eps = DBL_EPSILON * fmax(5.0, fabs((var.pf()).amp(0)));
  if (hbin < eps) hbin = 0.0;
  fout << "  実測平均HBI:全日=" << hbia << " 昼間=" << hbid;
  fout << " 夜間=" << hbin << " 昼間+夜間=" << hbid + hbin << "\n";
  hbia = var.pfhbi(0.0, (var.pf()).t(0), ha);
  hbid = var.pfhbi(opt.td(), opt.tn(), hd);
  hbin = var.pfhbi(0.0, (var.pf()).t(0), hn) - var.pfhbi(opt.td(), opt.tn(), hn);
  if (hbin < eps) hbin = 0.0;
  fout << "  周期関数HBI:全日=" << hbia << " 昼間=" << hbid;
  fout << " 夜間=" << hbin << " 昼間+夜間=" << hbid + hbin << "\n";
}

void pregclass::put_dw(ostream& fout, const option& opt, const idd& id, varclass& var) const
// ダービン・ワトソン比を出力する
{
  double e1 = var.mean(0) - (var.pf()).fun(opt.t(0));
  double sume2 = e1 * e1, sumed = 0.0, epre = e1;
  for (size_t j = 1; j < id.ni(); ++j)
  {
    double ei = var.mean(j) - (var.pf()).fun(opt.t(j));
    double ed = ei - epre;
    sume2 += ei * ei;  sumed += ed * ed;
    epre = ei;
  }
  double dw = sumed / sume2;
  fout << "\n・ダービン・ワトソン比:DW=" << dw << endl;
}

void pregclass::put_pregtab1(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sr, int dfr, double st, int dft) const
// 一元配置型周期回帰分散分析表を出力する
{
  string tstr("一元配置型周期分散分析表(periodic ANOVA table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << setw(62 + npad) << tstr;
  fout << "\n";
  fout << "要因              平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, sp, dfp, sr, dfr);
  ios::fmtflags flags0 = fout.flags(); // 現在の書式フラッグを保存する
  tstr = "周期回帰";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
  fout << setw(10) << dfp << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sl, dfl, sr, dfr);
  tstr = "ズレ";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
  fout << setw(10) << dfl << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, sa, dfa, sr, dfr);
  tstr = "時期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  tstr = "残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "全体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft;
  fout << endl;
}

void pregclass::put_pregtab2(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sb, int dfb, double sr, int dfr, double st, int dft) const
// 二元配置型周期回帰分散分析表を出力する
{
  string tstr("二元配置型周期分散分析表(periodic ANOVA table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << setw(62 + npad) << tstr;
  fout << "\n";
  fout << "要因              平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, sp, dfp, sr, dfr);
  tstr = "周期回帰";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
  fout << setw(10) << dfp << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sl, dfl, sr, dfr);
  tstr = "ズレ";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
  fout << setw(10) << dfl << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, sa, dfa, sr, dfr);
  tstr = "時期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  calvfp(v, f, pr, sb, dfb, sr, dfr);
  tstr = "個体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sb;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  tstr = "残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "全体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft;
  fout << endl;
}

void pregclass::plot_ngp(option& opt, idd& id, varclass& var) const
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("mse"))           // 平均値±SE
  {
    fout << "-s1 -dmark,00 -o300\n";
    plot_ngpid(fout, id);
    plot_ngpmse(fout, opt, id, var);
    delml = true;
  }
  if (opt.ngp("prc"))           // 周期回帰曲線
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80\n";
    plot_ngpid(fout, id);
    plot_ngpprc(fout, var.pf());
  }
}

void pregclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%時期項目  1:" << id.item(0) << "\n";
  fout << "%  :\n";
  fout << "%時期項目" << setw(3) << id.ni();
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void pregclass::plot_ngpmse(ofstream& fout, const option& opt, const idd& id, const varclass& var) const
// Ngraph用時期別平均±SEを出力する
{
  fout << "%平均±SE(SEを描く時は-derrorbar_y -w40 -o500)\n";
  for (size_t j = 0; j < id.ni(); ++j)
  {
    fout << opt.t(j) << " " << var.mean(j) << " " << -var.se(j) << " " << var.se(j) << "\n";
  }
}

void pregclass::plot_ngpprc(ofstream& fout, const pfun<double>& pf) const
// Ngraph用周期回帰曲線を出力する
{
  fout << "%周期回帰曲線\n";
  constexpr uint RDdef = 48;    // 曲線の分解数
  double dt = pf.t(0) / (double)RDdef;
  double t = 0.0;
  for (uint i = 0; i <= RDdef; ++i, t += dt)
  { fout << t << " " << pf.fun(t) << "\n"; }
}

int pregclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー！");
  return nc;
}

void pregclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var(id.ni());        // 変数オブジェクト
  if (var.cal(opt, datalst) < 1) return; // 計算できず
  if (opt.res("res"))
  {
    put_resid(cout, id);
    if (opt.res("stat")) put_stat(cout, opt, id, var);
    if (opt.res("reg")) put_pstat(cout, opt, id, var);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(opt, id, var);
}

int main(int argc, char* argv[])
// メインルメーチン
{
  pregclass preg;               // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) preg.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { preg.get(argv[i], opt, id, datalst);  fok = false; }
    }
    preg.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
