// sspect.C Version 7.41  Copyleft (c) 杉本典夫
// Time-stamp: <2023-12-15 11:30:21 tac>
//
// 時間項目と実数項目の層別周期解析を行い、スペクトルをデータファイル形式で出力する。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W sspect sspect.C -I/usr/local/include
//              -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "sspect  Version 7.41         Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:sspect  [<オプション> …]  [<データファイル> …]",
  "機能:時間項目と実数項目の層別周期解析を行い、",
  "     スペクトルをデータファイル形式で出力する。",
  "<オプション>",
  "  -dh=<データの時間間隔>:データの時間間隔を指定する。",
  "    オプション省略時:-dh=1",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込む。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示する。(-h:-h=man)",
  "  -ns=<スペクトル数>:出力するスペクトル数を指定する。",
  "    オプション省略時:-ns=10  -ns=0:全スペクトル出力",
  "  -T=<基本周期>:基本周期(1周期の時間)を指定するでござる。",
  "    オプション省略時:-T=300"
};

constexpr int T0def = 300;      // 省略時基本周期

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  int vdh;                      // データの時間間隔
  int vns;                      // 出力するスペクトル数
  int vt0;                      // 基本周期
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  int dh() const                // データの時間間隔
  { return vdh; }
  int ns() const                // 出力するスペクトル数を返す
  { return vns; }
  int t0() const                // 基本周期を返す
  { return vt0; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
};

option::option(const char* c_pname) : vpath(), vpname(), vdh(1), vns(10), vt0(T0def)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dh")           // データの時間間隔の指定
  {
    char* ptr = nullptr;
    vdh = strtol(opt[1].c_str(), &ptr, 0);
    if (vdh < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)vt0);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ns")      // 出力するスペクトル数の指定
  {
    char* ptr = nullptr;
    vns = strtol(opt[1].c_str(), &ptr, 0);
    int nsmax = vt0 / vdh / 2;
    if (vns < 0 || nsmax < vns) err = badopt(vpname, optstr, USAGE, 1.0, (double)nsmax);
  }
  else if (opt[0] == "T")       // 基本周期の指定
  {
    char* ptr = nullptr;
    vt0 = strtol(opt[1].c_str(), &ptr, 0);
    if (vt0 < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

class fft                       // FFTクラス
{
private:
  int tpre;                      // 直前の時間データ
  list<pair<int, double>> vylst; // データリスト
public:
  fft() : tpre(0), vylst()      // デフォールトコンストラクタ
  {}
  fft(int t0, double y0);       // コンストラクタ
  fft(const fft& fft0) : tpre(fft0.tpre), vylst(fft0.vylst)
  // コピーコンストラクタ
  {}
  virtual ~fft()                // デストラクタ
  {}
  const fft& operator=(const fft& fft0); // コピー代入演算子
  void clear();
  // データをクリアする
  int sigmay(int t, double y);  // データの累積集計
  const list<pair<int, double>>& ylst() const // データリストを返す
  { return vylst; }
};

fft::fft(int t0, double y0) : tpre(t0), vylst() // コンストラクタ
{
  vylst.push_back(make_pair(1, y0));
}

const fft& fft::operator=(const fft& fft0) // コピー代入演算子
{
  if (&fft0 == this) return *this; // 自分自身への代入を防ぐ
  tpre = fft0.tpre;  vylst = fft0.vylst;
  return *this;
}

void fft::clear()               // データをクリアする
{
  tpre = 0;  vylst.clear();
}

int fft::sigmay(int t, double y) // データの累積集計
{
  if (t <= tpre)
  { ++(vylst.back().first);  vylst.back().second += y; }
  else
  {
    vylst.push_back(make_pair(1, y));
    tpre = t;
  }
  return vylst.back().first;    // データ数を返す
}

using mapsf = map<string, fft>; // 群別FFTマップの別名定義
using mapsf_itr = mapsf::iterator; // 反復子の別名定義
using mapsf_citr = mapsf::const_iterator; // const反復子の別名定義

class sspectclass                // sspectメインクラス
{
private:
  string fname;                 // ファイル名
  void put_iddns(ostream& fout, const option& opt, idd& id) const;
  // データ識別情報をデータファイル形式で出力する
  void put_dns(ostream& fout, const option& opt, mapsf_citr fftitr) const;
  // 集計結果をデータファイル形式で出力する
  int cal_fftc(const option& opt, int& nd, vector<double>& abvec, const list<pair<int, double>>& ylst) const;
  // 基礎統計とFFTを計算し、FFT計算用データ数を返す
public:
  sspectclass() : fname()       // コンストラクタ
  {}
  ~sspectclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込む
  void put(option& opt, idd& id, ldata& datalst) const;
  // スペクトルを出力する
};

void sspectclass::put_iddns(ostream& fout, const option& opt, idd& id) const
// データ識別情報をデータファイル形式で出力する
{
  int p = opt.t0() / opt.dh();
  int ns = (opt.ns() > 0) ? opt.ns() : p / 2;
  const int Nitem = 4 + ns;     // 出力項目数
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  idd tid(id.title(), Nitem);
  tid.assign_item(0, id.item(0));
  tid.assign_item(1, id.item(2)+"-元データ数");
  tid.assign_item(2, id.item(2)+"-平均値");
  tid.assign_item(3, id.item(2)+"-標準偏差");
  for (int j = 0; j < ns; ++j)
  {
    ostringstream ostr;
    ostr << id.item(2) << "-スペクトル" << j + 1;
    tid.assign_item(4 + j, ostr.str());
  }
  tid.put_item(fout, formtag);
}

void sspectclass::put_dns(ostream& fout, const option& opt, mapsf_citr fftitr) const
// 集計結果をデータファイル形式で出力する
{
  int p = opt.t0() / opt.dh();
  int ns = (opt.ns() > 0) ? opt.ns() : p / 2;
  datac cdata(4 + ns);
  cdata.assign(0, fftitr->first);
  vector<double> abvec(p + 2);
  int nd = 0;
  int ny = cal_fftc(opt, nd, abvec, fftitr->second.ylst());
  double sd = (ny > 1) ? sqrt(abvec[p + 1] / (double)(ny - 1)) : 0.0;
  cdata.assign(1, nd);
  cdata.assign(2, abvec[0]);
  cdata.assign(3, sd);
  if (ny < p)                   // FFT計算不可能
  {
    for (int i = 0; i < ns; ++i) cdata.assign(4 + i, BLANKs);
  }
  else
  {
    int q = p / 2;
    vector<int> pcvec0(q);
    for (int k = 0; k < q; ++k) pcvec0[k] = k + 1;
    pfun pf((double)(opt.t0()), pcvec0, abvec);
    int dk = q / ns;
    double sapow = 0.0;
    int i = 4, na = 0;
    for (int k = 1; k <= q; ++k)
    {
      double dfk = (2 * k < p) ? 0.5 : 1.0;
      double apow = pf.amp(k) * pf.amp(k) * dfk;
      sapow += apow;  ++na;
      if (dk > 0 && na == dk)
      {
        cdata.assign(i, sapow / (double)na);
        sapow = 0.0;  na = 0;  ++i;
      }
    }
    if (na > 0) cdata.assign(i, sapow / (double)na);
  }
  cdata.put(fout);
}

int sspectclass::cal_fftc(const option& opt, int& nd, vector<double>& abvec, const list<pair<int, double>>& ylst) const
// 基礎統計とFFTを計算し、FFT計算用データ数を返す
{
  int p = opt.t0() / opt.dh();
  int q = p / 2;
  double wt = (double)(opt.dh()) * 2.0 * M_PI / (double)p;
  abvec.assign(p + 2, 0.0);
  list<pair<int, double>>::const_iterator yitr = ylst.begin();
  int ny = 0;
  nd = 0;
  for (int j = 0; j < p && yitr != ylst.end(); ++j, ++yitr, ++ny)
  {
    nd += yitr->first;          // 元データ数
    double y = (yitr->first > 1) ? yitr->second / (double)(yitr->first) : yitr->second;
    for (int k = 1; k <= q; ++k)
    {
      int i = 2 * k;
      double tk = wt * (double)(k * j);
      abvec[i - 1] += y * cos(tk);
      if (i < p) abvec[i] += y * sin(tk);
    }
    abvec[0] += y;  abvec[p + 1] += y * y;
  }
  for (int k = 1; k <= q; ++k)
  {
    int i = 2 * k;
    if (i == p) abvec[i - 1] /= (double)p;
    else { abvec[i - 1] /= (double)p * 0.5;  abvec[i] /= (double)p * 0.5; }
  }
  const double ABeps = DBL_EPSILON * fmax(5.0, abvec[p + 1]);
  abvec[p + 1] -= abvec[0] * abvec[0] / (double)ny; // FFT計算用データの平方和(St)
  if (abvec[p + 1] < ABeps) abvec[p + 1] = 0.0;
  abvec[0] /= (double)ny;       // FFT計算用データの平均(My)
  return ny;                    // FFT計算用データ数を返す
}

int sspectclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (id.ni() != 3) throw runtime_error("項目は3個です！");
  return nc;
}

void sspectclass::put(option& opt, idd& id, ldata& datalst) const
// スペクトルを出力する
{
  put_iddns(cout, opt, id);
  if (id.nc() < 1) return;      // データなし
  mapsf fftmap;                 // FFTクラスマップ
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    if (ditr->sdata(0) != BLANKs) // 層別コードあり
    {
      mapsf_itr fftitr = fftmap.find(ditr->sdata(0));
      if (fftitr != fftmap.end()) // 既存コード
      {
        double t = ditr->fdata(1), y = ditr->fdata(2);
        if (t != BLANKf && y != BLANKf)
        {
          int tint = (int)t / opt.dh();
          fftitr->second.sigmay(tint, y);
        }
      }
      else                      // 新規コード
      {
        double t = ditr->fdata(1), y = ditr->fdata(2);
        if (t != BLANKf && y != BLANKf)
        {
          int tint = (int)t / opt.dh();
          fft tfft(tint, y);
          fftmap.insert(mapsf::value_type(ditr->sdata(0), tfft));
        }
      }
    }
  }
  for (mapsf_citr fftitr = fftmap.begin(); fftitr != fftmap.end(); ++fftitr) put_dns(cout, opt, fftitr);
}

int main(int argc, char* argv[]) // メインルーチン
{
  sspectclass sspect;           // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) sspect.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { sspect.get(argv[i], opt, id, datalst);  fok = false; }
    }
    sspect.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
