// tprat.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2023-12-15 11:32:47 tac>
//
// 個体別、平均値、周期回帰曲線によるT/P比の集計を行う。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W tprat tprat.C -I/usr/local/include
//               -L/usr/local/lob -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H,mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "tprat  Version 7.42          Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:tprat  [<オプション> …]  [<データファイル> …]",
  "機能:個体別、平均値、周期回帰曲線によるT/P比を集計します。",
  "<オプション>",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込みます。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示します。(-h:-h=man)",
  "  -k=<番号>[,<番号>,…]:周期回帰式に組み込む周期成分番号を指定します。",
  "    オプション省略時:-k=1,2",
  "  -pm={0|1|2}:周期解析の手法を指定します。",
  "    0:平均値を用いる単純法  1:一元配置法  2:二元配置法(オプション省略時)",
  "  -T=<基本周期>:基本周期(1周期の時間)を指定します。",
  "  -t=<時間1>,<時間2>[,<時間3>,…]:時期項目の時間を指定します。",
  "    オプション省略時:-t=0,T/時期数,…,(時期数-1)×T/時期数",
  "  -tt={-|<trough時間>}:trough時間を指定します。",
  "    -:指定せず(オプション省略時)  <trough時間>:trough時間を固定して計算"
};

const string Title = "=== 周期回帰分析(periodic regression analysis) ==="; // 表題
constexpr double T0def = 24.0;  // 省略時基本周期

using vecstatv = vector<statv<double>>; // 基本統計量ベクトルの別名定義

enum pmtype                     // 周期解析の種類
{
  pmnon,                        // なし
  pm0,                          // 平均値を用いる単純法
  pm1,                          // 一元配置法
  pm2,                          // 二元配置法
  pmetc                         // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  lstr pclst;                   // 周期成分番号リスト
  vecint vpcvec;                // 周期成分番号配列
  int vpcmax;                   // 周期成分番号最大値
  pmtype vpm;                   // 周期解析の種類
  lstr tlst;                    // 項目時間リスト
  vecdbl vtvec;                 // 項目時間ベクトル
  double vt0;                   // 基本周期
  double vtt;                   // trough時間
  int vtidx;                    // trough時間の項目インデックス
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  int m() const                 // 周期成分数を返す
  { return vpcvec.size(); }
  int pcn(size_t k) const       // 周期成分番号を返す
  { return (k < vpcvec.size()) ? vpcvec[k] : 0; }
  const vecint& pcvec() const   // 周期成分番号ベクトルを返す
  { return vpcvec; }
  int pcmax() const             // 周期成分番号最大値を返す
  { return vpcmax; }
  pmtype pm() const             // 周期解析の種類を返す
  { return vpm; }
  double t(size_t idx) const    // 項目時間を返す
  { return (idx < vtvec.size()) ? vtvec[idx] : vt0; }
  double t0() const             // 基本周期を返す
  { return vt0; }
  double tt() const             // trough時間を返す
  { return vtt; }
  int tidx() const              // trough時間の項目インデックスを返す
  { return vtidx; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int p);             // オプションをチェックする
};

option::option(const char* c_pname) : vpath(), vpname(), pclst(), vpcvec(), vpcmax(-1), vpm(pm2), tlst(), vtvec(), vt0(T0def), vtt(DBL_MAX), vtidx(-1)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "k")       // 周期成分番号の指定
  {
    if (strsplit(pclst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pm")      // 周期解析の手法指定
  {
    if (opt[1].empty() || opt[1] == "2") vpm = pm2;
    else if (opt[1] == "0") vpm = pm0;
    else if (opt[1] == "1") vpm = pm1;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "T")       // 基本周期の指定
  {
    vt0 = strtodf(opt[1]);
    if (vt0 == BLANKf || vt0 <= 0.0) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "t")       // 項目時間の指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "tt")      // trough時間の指定
  {
    double fdata = 0.0;
    if (opt[1].empty()) vtt = BLANKf;
    else if (opt[1] == "-") vtt = DBL_MAX;
    else vtt = fdata = strtodf(opt[1]);
    if (fdata == BLANKf || vtt < 0.0 || 24.0 <= vtt) err = badopt(vpname, optstr, USAGE, 0.0, 24.0);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(int p)        // オプションをチェックする
{
  if (p < 2) return disperr(vpname, "時期項目数が少なすぎます！");
  int q = p / 2;                // 周期成分最大数
  vtvec.resize(p);
  double dt = vt0 / (double)p;
  for (int j = 0; j < p; ++j) vtvec[j] = (double)j * dt;
  if (settvec(vtvec, tlst, p) < 0) return disperr(vpname, "項目時間が間違っています！");
  if (vtt != DBL_MAX)           // tidxの設定
  {
    for (int j = 0; vtidx == -1 && j < p; ++j)
    {
      if (vtvec[j] >= vtt) vtidx = j;
    }
  }
  if (pclst.empty())            // 周期成分指定がない時
  {
    vpcvec.resize(2);
    vpcvec[0] = 1;  vpcvec[1] = 2;  vpcmax = 2;
  }
  else if ((vpcmax = setpcvec(vpcvec, pclst, q)) < 0)
  { return disperr(vpname, "周期成分指定が間違っています！"); }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  int vnp;                      // 時期項目数
  vecstatv ystatvec;            // 時期項目の基本統計量ベクトル
  vecstatv tpstatvec;           // 個体別T/P比の基本統計量ベクトル
  int vnb;                      // 観察個体数
  double vsb;                   // 観察個体平方和
  pfun<double> vpf;             // 周期関数オブジェクト
  mat<double> smat;             // 周期回帰計算用積和行列
  double vst;                   // 全体平方和
  int vdft;                     // 全体自由度
  double vsa;                   // 群差平方和
  int vdfa;                     // 群差自由度
  int vdfb;                     // 回帰自由度
  double vsp;                   // 時期平方和
  int vdfp;                     // 時期自由度
  double vsl;                   // ズレ平方和
  int vdfl;                     // ズレ自由度
  double vsr;                   // 残差平方和
  int vdfr;                     // 残差自由度
  int cal_tprat(const option& opt);
  // 基礎統計・周期回帰式を計算する
public:
  varclass(size_t ni) : vnp(ni), ystatvec(ni + 1), tpstatvec(2), vnb(0), vsb(0.0), vpf(), smat(), vst(0.0), vdft(0), vsa(0.0), vdfa(0), vdfb(0), vsp(0.0), vdfp(0), vsl(0.0), vdfl(0), vsr(0.0), vdfr(0) // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int np() const                // 時期項目数を返す
  { return vnp; }
  int nb() const                // 観察個体数を返す
  { return vnb; }
  int n(size_t idx) const       // 時期項目の例数を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].n() : 0; }
  double mean(size_t idx) const // 時期項目の平均を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].mean() : 0.0; }
  double vr(size_t idx) const   // 時期項目の不偏分散を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].vr() : 0.0; }
  double se(size_t idx) const   // 時期項目の標準誤差を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].se() : 0.0; }
  int tpn(size_t idx) const     // 個体別T/P比の例数を返す
  { return (idx < tpstatvec.size()) ? tpstatvec[idx].n() : 0; }
  double tpmean(size_t idx) const // 個体別T/P比の平均を返す
  { return (idx < tpstatvec.size()) ? tpstatvec[idx].mean() : 0.0; }
  double tpvr(size_t idx) const // 個体別T/P比の不偏分散を返す
  { return (idx < tpstatvec.size()) ? tpstatvec[idx].vr() : 0.0; }
  double sb() const             // 観察個体平方和を返す
  { return vsb; }
  double ss(size_t iidx, size_t jidx) const // 積和行列の要素を返す
  { return (iidx < smat.r() && jidx < smat.c()) ? smat(iidx, jidx) : 0.0; }
  pfun<double>& pf()            // 周期関数オブジェクトを返す
  { return vpf; }
  double st() const             // 全体平方和を返す
  { return vst; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
  double sa() const             // 群差平方和を返す
  { return vsa; }
  int dfa() const               // 群差自由度を返す
  { return vdfa; }
  int dfb() const               // 回帰自由度を返す
  { return vdfb; }
  double sp() const             // 時期平方和を返す
  { return vsp; }
  int dfp() const               // 時期自由度を返す
  { return vdfp; }
  double sl() const             // ズレ平方和を返す
  { return vsl; }
  int dfl() const               // ズレ自由度を返す
  { return vdfl; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  int cal(const option& opt, ldata& datalst); // データを集計して統計量を求める
};

int varclass::cal_tprat(const option& opt)
// 基礎統計・周期回帰式を計算する
{
  int p = vnp, dim = 2 * opt.m() + 2, np = 0;
  int yidx = dim - 1;           // yのインデックス
  vecdbl xyvec(dim);
  const double W = 2.0 * M_PI / opt.t0();
  xyvec[0] = 1.0;  vsa = vst = 0.0;
  for (int j = 0; j < p; j++)
  {
    int n = ystatvec[j].n();
    if (n > 0)
    {
      double me = ystatvec[j].mean();
      xyvec[yidx] = me;  ystatvec[p] += ystatvec[j];
      for (int k = 0; k < opt.m(); ++k)
      {
        int i = 2 * (k + 1);
        double tk = W * opt.pcn(k) * opt.t(j);
        xyvec[i - 1] = cos(tk);
        xyvec[i] = (2 * opt.pcn(k) < p) ? sin(tk) : 0.0;
      }
      int r = (opt.pm() == pm0) ? 1 : n;
      for (int i = 0; i < dim; ++i)
      {
        double x = r * xyvec[i];
        for (int k = i; k < dim; ++k) smat.add(i, k, x * xyvec[k]);
      }
      ++np;  vsa += n * me * me;
    }
  }
  if (np < yidx) return disperr(opt.pname(), "時期項目が少なすぎます！");
  for (int i = 0; i < yidx; i++)
  {
    for (int j = i + 1; j < dim; j++) smat.assign(j, i, smat(i, j));
  }
  smat.sweep(0);
  if (opt.pm() != pm0) smat.assign(yidx, yidx, ystatvec[p].ss());
  vst = smat(yidx, yidx);
  vdfa = np - 1;
  vdft = (opt.pm() == pm0) ? vdfa : ystatvec[p].n() - 1;
  for (int j = 1; j < yidx; ++j) smat.sweep(j);
  double cf = ystatvec[p].n() * ystatvec[p].mean() * ystatvec[p].mean();
  vsa -= cf;  vsb = vsb / (double)p - cf;  vsp = vst - smat(yidx, yidx);
  vsl = vsa - vsp;
  const double Seps = DBL_EPSILON * fmax(5.0, vst);
  if (vsa < Seps) vsa = 0.0;
  if (vsb < Seps) vsb = 0.0;
  if (vsp < Seps) vsp = 0.0;
  if (vsl < Seps) vsl = 0.0;
  vdfp = 2 * (int)(opt.m());
  if (2 * opt.pcmax() >= p) --vdfp;
  vdfl = vdfa - vdfp;  vdfb = vnb - 1;  vdfr = vdft - vdfp;
  if (vdfl < 1) vsl = 0.0;
  vsr = vst - vsp;
  if (opt.pm() == pm1)
  { vsr = vst - vsa;  vdfr = vdft - vdfa; }
  else if (opt.pm() == pm2)
  { vsr = vst - vsa - vsb;  vdfr = vdft - vdfa - vdfb; }
  if (vdfr < 1 || vsr < Seps) vsr = 0.0;
  return np;                    // 有効な時期項目数を返す
}

int varclass::cal(const option& opt, ldata& datalst)
// データを集計して統計量を求める
{
  vecdbl fdatavec(vnp);
  for (size_t j = 0; j < ystatvec.size(); ++j) ystatvec[j].clear();
  vnb = 0;  vsb = 0.0;  tpstatvec[0].clear();  tpstatvec[1].clear();
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    bool blank = false;
    double trough = -DBL_MAX, peak = DBL_MAX;
    for (size_t j = 0; j < fdatavec.size(); ++j)
    {
      fdatavec[j] = ditr->fdata(j);
      if (fdatavec[j] == BLANKf)
      { blank = true; }
      else
      {
        if (fdatavec[j] > trough) trough = fdatavec[j];
        if (fdatavec[j] < peak) peak = fdatavec[j];
      }
    }
    if (blank == false && peak != DBL_MAX) // trouphとpeakが計算できる時
    {
      double tp = (trough < 0.0) ? trough / peak : 0.0;
      tpstatvec[0].sigma(tp);
      if (opt.tidx() >= 0 && fdatavec[opt.tidx()] != BLANKf)
      {
        tp = (fdatavec[opt.tidx()] < 0.0) ? fdatavec[opt.tidx()] / peak : 0.0;
        tpstatvec[1].sigma(tp);
      }
    }
    if (blank == false || opt.pm() != pm2) // 二元配置型ではない時
    {
      double sum = 0.0;
      for (size_t j = 0; j < fdatavec.size(); ++j)
      {
        if (fdatavec[j] != BLANKf)
        { ystatvec[j].sigma(fdatavec[j]);  sum += fdatavec[j]; }
      }
      ++vnb;  vsb += sum * sum;
    }
  }
  int dim = 2 * opt.m() + 2;
  smat.resize(dim, dim);
  int np = cal_tprat(opt);
  vecdbl abvec(dim - 1);
  for (int j = 0; j < dim - 1; ++j) abvec[j] = smat(j, dim - 1);
  pfun<double> pf(opt.t0(), opt.pcvec(), abvec);
  vpf = pf;
  return np;                    // 有効な時期項目数を返す
}

class tpratclass                 // tpratメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 時期別基礎統計量を出力する
  void put_tpratfun(ostream& fout, const option& opt, varclass& var) const;
  // 周期関数パラメターを出力する
  void put_tpratio(ostream& fout, const option& opt, const idd& id, varclass& var) const;
  // T/P比を出力する
  void put_tprattab1(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sr, int dfr, double st, int dft) const;
  // 一元配置型周期回帰分散分析表を出力する
  void put_tprattab2(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sb, int dfb, double sr, int dfr, double st, int dft) const;
  // 二元配置型周期回帰分散分析表を出力する
public:
  tpratclass() : fname()        // コンストラクタ
  {}
  ~tpratclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void tpratclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "時期項目  1:" << id.item(0) << "\n";
  fout << "  :\n";
  fout << "時期項目" << setw(3) << id.ni();
  fout << ":" << id.item(id.ni() - 1) << "\n";
}

void tpratclass::put_stat(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 時期別基礎統計量を出力する
{
  constexpr int LENline = 77;
  int p = id.ni();
  fout << "\n時間 t   :基礎統計量\n";
  putline(fout, LENline);
  for (int j = 0; j < p; ++j)
  {
    int prec0 = fout.precision(2); // 小数点以下の精度を指定する
    fout.setf(ios::fixed);
    fout << setw(9) << opt.t(j) << setprecision(prec0);
    fout.unsetf(ios::fixed);
    fout << ":";  putmsd(fout, var.n(j), var.mean(j), var.vr(j));
  }
  putline(fout, LENline);
  fout << "全体     :";  putmsd(fout, var.n(p), var.mean(p), var.vr(p));
}

void tpratclass::put_tpratfun(ostream& fout, const option& opt, varclass& var) const
// 周期関数パラメターを出力する
{
  fout << "\n・周期回帰式(periodic regression equation)\n";
  fout << "  y=a0/2+Σ{ak・cos(ωk・t)+bk・sin(ωk・t)}=A0+ΣAk・cos(ωk・t-θk)\n";
  fout << "    k=1,…,p:周期成分  A0:メサー(定数)  Ak:振幅  ωk=360°・k/T:角周波数\n";
  fout << "    θk:位相  T=" << (var.pf()).t(0) << ":基本周期  T/k:周期  f=1/T:基本周波数  fk=k/T:周波数\n";
  fout << "\n周期成分      振幅  位相(度)  位相時間            回帰係数    角周波数    周期\n";
  fout << "   k            Ak      θk°  θk/ωk          ak          bk     ωk     T/k\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  fout << "   0  : " << setw(10) << (var.pf()).amp(0) << "(メサー)\n";
  double deg = 180.0 / M_PI;
  for (size_t k = 1; k <= (var.pf()).m(); ++k)
  {
    fout << setw(4) << (var.pf()).pcn(k);
    fout << "  : " << setw(10) << (var.pf()).amp(k) << setw(10) << (var.pf()).acr(k) * deg;
    fout << setw(10) << (var.pf()).acr(k) / (var.pf()).w(k) << setw(12) << (var.pf()).a(k);
    fout << setw(12) << (var.pf()).b(k) << setw(8) << (var.pf()).w(k) * deg;
    fout << setw(8) << (var.pf()).t(k) << "\n";
  }
  putline(fout, LENline);
  double r2 = (var.st() > 0.0) ? var.sp() / var.st() : 0.0;
  if (opt.pm() == pm0)
  {
    fout << "周期回帰式の重寄与率R2乗=" << r2 << "  重相関係数R=" << sqrt(r2) << endl;
  }
  else
  {
    r2 = (var.sa() > 0.0) ? var.sp() / var.sa() : 0.0;
    fout << "平均値に対する周期回帰式の重寄与率R2乗=" << r2 << "  重相関係数R=" << sqrt(r2) << "\n";
    r2 = (var.st() > 0.0) ? var.sp() / var.st() : 0.0;
    fout << "データに対する周期回帰式の重寄与率R2乗=" << r2 << "  重相関係数R=" << sqrt(r2) << endl;
  }
}

void tpratclass::put_tpratio(ostream& fout, const option& opt, const idd& id, varclass& var) const
// T/P比を出力する
{
  fout << "\n";
  if (var.tpn(0) > 0)
  {
    fout << "・個体別T/P比\n";
    fout << "  Tmax:";  putmsd(fout, var.tpn(0), var.tpmean(0), var.tpvr(0));
  }
  if (var.tpn(1) > 0)
  {
    fout << "  t=" << setw(2) << opt.t(opt.tidx()) << ":";
    putmsd(fout, var.tpn(1), var.tpmean(1), var.tpvr(1));
  }
  double trough = -DBL_MAX, tt = 0.0, peak = DBL_MAX, pt = 0.0, tp = 0.0;
  for (size_t j = 0; j < id.ni(); ++j)
  {
    if (var.mean(j) > trough)
    { trough = var.mean(j);  tt = opt.t(j); }
    if (var.mean(j) < peak)
    { peak = var.mean(j);  pt = opt.t(j); }
  }
  if (peak != DBL_MAX)
  {
    tp = (trough < 0.0) ? trough / peak : 0.0;
    fout << "・平均値によるT/P比\n";
    fout << "  T(最大値使用) :T=" << trough << "(t=" << tt << ")  p=" << peak;
    fout << "(t=" << pt << ")  T/P=" << tp << "\n";
    if (opt.tidx() >= 0 && var.n(opt.tidx()) > 0)
    {
      trough = var.mean(opt.tidx());  tt = opt.t(opt.tidx());
      tp = (trough < 0.0) ? trough / peak : 0.0;
      fout << "  T(t=" << tt << "の値使用):T=" << trough << "(t=" << tt;
      fout << ")  p=" << peak << "(t=" << pt << ")  T/P=" << tp << "\n";
    }
  }
  trough = -DBL_MAX;  peak = DBL_MAX;
  constexpr uint Dev = 720;     // 関数サーチ時の分解数
  double dt = (var.pf()).t(0) / (double)Dev;
  double t = 0.0;
  for (uint i = 0; i <= Dev; ++i, t += dt)
  {
    double y = (var.pf()).fun(t);
    if (y > trough) { trough = y;  tt = t; }
    if (y < peak) { peak = y;  pt = t; }
  }
  if (peak != DBL_MAX)
  {
    tp = (trough < 0.0) ? trough / peak : 0.0;
    fout << "・周期曲線によるT/P比\n";
    fout << "  T(最大値使用) :T=" << trough << "(t=" << tt << ")  p=" << peak;
    fout << "(t=" << pt << ")  T/P=" << tp << "\n";
    if (opt.tt() != DBL_MAX)
    {
      trough = (var.pf()).fun(opt.tt());  tt = opt.tt();
      tp = (trough < 0.0) ? trough / peak : 0.0;
      fout << "  T(t=" << tt << "の値使用):T=" << trough << "(t=" << tt;
      fout << ")  p=" << peak << "(t=" << pt << ")  T/P=" << tp << "\n";
    }
  }
}

void tpratclass::put_tprattab1(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sr, int dfr, double st, int dft) const
// 一元配置型周期回帰分散分析表を出力する
{
  string tstr("一元配置型周期分散分析表(periodic ANOVA table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << setw(62 + npad) << tstr;
  fout << "\n";
  fout << "要因              平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  double v = 0.0, f = 0.0, p = 0.0;
  double vr = calvfp(v, f, p, sp, dfp, sr, dfr);
  tstr = "周期回帰";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
  fout << setw(10) << dfp << setw(14) << v << setw(12) << f << setw(14) << p;
  fout << pind(p) << "\n";
  calvfp(v, f, p, sl, dfl, sr, dfr);
  tstr = "ズレ";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
  fout << setw(10) << dfl << setw(14) << v << setw(12) << f << setw(14) << p;
  fout << pind(p) << "\n";
  putline(fout, LENline);
  calvfp(v, f, p, sa, dfa, sr, dfr);
  tstr = "時期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << p;
  fout << pind(p) << "\n";
  tstr = "残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "全体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft << endl;
}

void tpratclass::put_tprattab2(ostream& fout, double sp, int dfp, double sl, int dfl, double sa, int dfa, double sb, int dfb, double sr, int dfr, double st, int dft) const
// 二元配置型周期回帰分散分析表を出力する
{
  string tstr("二元配置型周期分散分析表(periodic ANOVA table)");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << setw(62 + npad) << tstr;
  fout << "\n";
  fout << "要因              平方和    自由度    平均平方和         F値   有意確率p値\n";
  constexpr int LENline = 78;
  putline(fout, LENline);
  double v = 0.0, f = 0.0, p = 0.0;
  double vr = calvfp(v, f, p, sp, dfp, sr, dfr);
  tstr = "周期回帰";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
  fout << setw(10) << dfp << setw(14) << v << setw(12) << f << setw(14) << p;
  fout << pind(p) << "\n";
  calvfp(v, f, p, sl, dfl, sr, dfr);
  tstr = "ズレ";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sl;
  fout << setw(10) << dfl << setw(14) << v << setw(12) << f << setw(14) << p;
  fout << pind(p) << "\n";
  putline(fout, LENline);
  calvfp(v, f, p, sa, dfa, sr, dfr);
  tstr = "時期";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f << setw(14) << p;
  fout << pind(p) << "\n";
  calvfp(v, f, p, sb, dfb, sr, dfr);
  tstr = "個体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sb;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f << setw(14) << p;
  fout << pind(p) << "\n";
  tstr = "残差";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr;
  fout << "\n";
  putline(fout, LENline);
  tstr = "全体";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft << endl;
}

int tpratclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (id.ni() < 2) throw runtime_error("時期項目が2つ以上必要です！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー");
  return nc;
}

void tpratclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  put_id(cout, id);
  if (id.nc() < 1)              // データなし
  {
    cout << "\n" << endl;
    return;
  }
  varclass var(id.ni());        // 変数オブジェクト
  if (var.cal(opt, datalst) < 1) return; // 計算できず
  put_stat(cout, opt, id, var);
  put_tpratfun(cout, opt, var);
  put_tpratio(cout, opt, id, var);
  if (var.st() > 0.0)
  {
    cout << "\n";
    switch(opt.pm())
    {
    case pm0 : putatab(cout, "周期回帰", var.sp(), var.dfp(), var.sr(), var.dfr(), var.st(), var.dft());
      break;
    case pm1 : put_tprattab1(cout, var.sp(), var.dfp(), var.sl(), var.dfl(), var.sa(), var.dfa(), var.sr(), var.dfr(), var.st(), var.dft());
      break;
    default : put_tprattab2(cout, var.sp(), var.dfp(), var.sl(), var.dfl(), var.sa(), var.dfa(), var.sb(), var.dfb(), var.sr(), var.dfr(), var.st(), var.dft());
    }
  }
  cout << "\n" << endl;
}

int main(int argc, char* argv[]) // メインルメーチン
{
  tpratclass tprat;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) tprat.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { tprat.get(argv[i], opt, id, datalst);  fok = false; }
    }
    tprat.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
